/*---------------------------------------------------------------------------
  API.C - DOS library for direct access to RocketPort hardware.

9-26-97 PCI updates.
  *BugFix, PCI operation, if IRQ is used, was calling EOI
   routine for ISA board.
  *BugFix, PCI operation, board type not being set in board
   structure if PCI, added.
  *Update PCI board ID list.

2-12-97 Header list(macro) Baud rates not working in aaOpen() and
  ReConfigure() call.  Fixed by replacing sSetBaudRate() call with
  aaSetBaud() call which translates the list-style entries to
  actual baud-rates.

10-23-96  Finish testing sample programs with new low-level code,
  add PCI board support to API.C and term.c sample program.

8-18-96 Use updated rckt.h low-level routines(pci-support), update baud
 rate handling to accept actual baud rates as well as an index value.
 General re-organization of things. - kpb

7-28-96 Add aaSetConfig routine as alternative to use of configuration file. - kpb

4-02-96 Added aaGetRxStatusWord() for someone who needed to detect break
  condition - kpb.

4-02-96 Added aaPollHandler() and minor changes to allow running without IRQ.

 Copyright 1993-97 Comtrol Corporation.  All rights reserved.  Proprietary
 information only for development or use with Comtrol products.
|----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include <dos.h>
#ifdef BORLANDC
#include <alloc.h>
#endif
#if defined(MSC) || defined(WATCOMC)
#include <malloc.h>
#endif
#include "rcktldef.h"
#include "rckt.h"
#include "api.h"
#include "pcibios.h"

// #define TRY_EVENT_IRQS 1

/* Configuration information structure for one controller */
typedef struct
{
   WORD MudbacIO;                    /* I/O address of MUDBAC */
   WORD MaxPorts;                    /* Max number of ports on board */
   int  BusType;                     // BUS_ISA, BUS_PCI
   WORD AiopIO[AIOP_CTL_SIZE];       /* I/O addresses of AIOPs */
   int StartDev;                       /* starting device number on this ctl */
                                       /*    -1 = ctl not installed */
   int NumChan;                        /* number channels on this controller */
} CFCTL_T;

/* Configuration information for all controllers */
typedef struct
{
   int Irq;                            /* IRQ number */
   int NumCtl;                         /* number of controllers in system */
   CFCTL_T CfCtl[CTL_SIZE];
} CF_T;

typedef struct                         /* Interrupt information structure */
{
   unsigned char Mask;                 /* mask value for 8259 mask register */
   WORD IntNum;
   WORD MaskIO;                /* I/O address of 8259 mask register */
} IRQ_T;

/* Device information structure */
typedef struct
{
   int Dev;                            /* device number */
   WORD Hnd;                         /* handle */
   WORD Status;                      /* device status */
   WORD DetectEn;                    /* detection enables */
   long BreakTime;                     /* time left on break signal */
   long CloseDelay;                    /* max Tx drain delay on close */
   CHANNEL_T Chan;                     /* channel structure */
} DEV_T;

CF_T aaCfg;                             /* configuration file data */
DEV_T *aaDevices[DEV_SIZE];         /* pointers to device structures */
int far aaCtlSize = CTL_SIZE;           /* number of supported controllers */
int far aaDevSize;                      /* number of supported devices */

int aaIrqSave;                         /* IRQ number save */
void (interrupt far *aaVectorSave)(void);  /* IRQ line interrupt vector save */
unsigned char aaMaskSave;              /* 8259 mask save */
WORD far aaEndOfIntIO;         /* 8259 I/O address for EOI */
unsigned long aaControllerTime;        /* controller time in micro seconds */
int aaPeriodicEventOn;                 /* TRUE if periodic event enabled */
int scan_for_mudback(WORD io_addr);
int scan_40hex_space(WORD io_addr);

// #define TEST_puts(s) puts(s)
#define TEST_puts(s) {}
/*******************************************************************************
  Define simple arrays that aid in the error checking for the API
  Library function parameters.
*******************************************************************************/
#define MAXDEV DEV_SIZE - 1
#define MAXNUMOTS 3
#define MAXNUMBAUDS COM_BAUD_230400 + 1
#define MAXNUMPARITYS 3
#define MAXNUMDATABS 2
#define MAXNUMSTOPBS 2

static WORD aaOpenTypeCheck[MAXNUMOTS]={COM_TX, COM_RX, COM_TX | COM_RX};

static unsigned char aaParityCheck[MAXNUMPARITYS]={COM_PAR_NONE,COM_PAR_EVEN,COM_PAR_ODD};

static unsigned char aaDataBCheck[MAXNUMDATABS]={COM_DATABIT_7,COM_DATABIT_8};

static unsigned char aaStopBCheck[MAXNUMSTOPBS]={COM_STOPBIT_1,COM_STOPBIT_2};


#define FLOWCHECK  (COM_FLOW_NONE | COM_FLOW_IS | COM_FLOW_IH | \
                   COM_FLOW_OS | COM_FLOW_OH | COM_FLOW_OXANY)

#define DETECTCHECK  (COM_DEN_RDA | COM_DEN_MDM)

#define MDMCTLCHECK  (COM_MDM_DTR | COM_MDM_RTS)

/* Array of baud rate divisors */
unsigned long aaBaud[22] =
{
   50,
   75,
   110,
   134,
   150,
   200,
   300,
   600,
   1200,
   1800,
   2400,
   3600,
   4800,
   7200,
   9600,
   19200L,
   38400L,
   57600L,
   76800L,
   115200L,
   230400L,
   460800L};

/* Array for converting IndID reg modem change bits into COM_MDM_XXX flags */
unsigned char aaModemChg[8] =
{
   0,COM_MDM_DSR,COM_MDM_CTS,COM_MDM_DSR|COM_MDM_CTS,COM_MDM_CD,
   COM_MDM_CD|COM_MDM_DSR,COM_MDM_CD|COM_MDM_CTS,COM_MDM_CD|COM_MDM_CTS|COM_MDM_DSR
};

#define DELTA_CD  0x04        /* CD change interrupt */
#define DELTA_CTS 0x02        /* CTS change interrupt */
#define DELTA_DSR 0x01        /* DSR change interrupt */


//---  Define the event function pointers.
void (*aaMdmChgEventP)(int,unsigned char,unsigned char);
void (*aaRxEventP)(int);
void (*aaPeriodicEventP)(void);

IRQ_T aaIrqInfo[] =                 /* IRQ information */
{
  {0xfe,0x8,0x21}, /* IRQ0 */
  {0xfd,0x9,0x21}, /* IRQ1 */
  {0xfb,0xa,0x21}, /* IRQ2 */
  {0xf7,0xb,0x21}, /* IRQ3 */
  {0xef,0xc,0x21}, /* IRQ4 */
  {0xdf,0xd,0x21}, /* IRQ5 */
  {0xbf,0xe,0x21}, /* IRQ6 */
  {0x7f,0xf,0x21}, /* IRQ7 */
  {0xfe,0x70,0xa1}, /* IRQ8 */
  {0xfd,0x71,0xa1}, /* IRQ9 */
  {0xfb,0x72,0xa1}, /* IRQ10 */
  {0xf7,0x73,0xa1}, /* IRQ11 */
  {0xef,0x74,0xa1}, /* IRQ12 */
  {0xdf,0x75,0xa1}, /* IRQ13 */
  {0xbf,0x76,0xa1}, /* IRQ14 */
  {0x7f,0x77,0xa1}  /* IRQ15 */
};

int config_setup = 0;

static int GetDevInfo(DEVINFO *pci_config);
static int FindDevs(int vendor_id,
                    int max_device_id,
                    int *num_pci_boards,
                    DEVINFO *pci_config);
#define MAX_DEV_ID    0x20      // our 4-port=1, 8-port=2, 16=3, 32-4.
#define COMTROL_ID    0x11FE    // Comtrols assigned vendor id
/*-----------------------------------------------------------------------
Function: aaFindPCIBoards
Purpose: Find all RocketPort PCI boards and return configuration information
  on each one in callers array of DEVINFO structures.
Call:
  aaFindPCIBoards(DeviceInfoArray, NumDevs);
      DEVINFO *DeviceInfoArray; // Caller should pass a pointer to an
         // array of four of these, we fill them in with board info we find
      int *NumDevs;  // The number of PCI RocketPort boards found in the
                     // system is returned here.
Return:   int: Zero is OK, Non-zero is error code.
|-------------------------------------------------------------------------*/
int aaFindPCIBoards(DEVINFO *DeviceInfoArray, int *NumDevs)
{
  
 int stat, i;
 unsigned int id;

  *NumDevs = 0;
  stat = pci_init();

  if (stat == 0)  // ok
  {
    stat = FindDevs(COMTROL_ID, MAX_DEV_ID, NumDevs, DeviceInfoArray);
    for (i=0; i<*NumDevs; i++)
    {
      GetDevInfo(&DeviceInfoArray[i]);  // get irq & addr for 1st board
      // boardio = DeviceInfoArray[i].Base_Addr;
      id = DeviceInfoArray[i].id;
      switch (id)
      {
        case PCI_DEVICE_32I: DeviceInfoArray[i].num_ports = 32; break;
        case PCI_DEVICE_8I: DeviceInfoArray[i].num_ports = 8; break;
        case PCI_DEVICE_16I: DeviceInfoArray[i].num_ports = 16; break;
        case PCI_DEVICE_4Q: DeviceInfoArray[i].num_ports = 4; break;
        case PCI_DEVICE_8O: DeviceInfoArray[i].num_ports = 8; break;
        case PCI_DEVICE_8RJ: DeviceInfoArray[i].num_ports = 8; break;

        case PCI_DEVICE_SIEMENS8 : DeviceInfoArray[i].num_ports = 8; break;
        case PCI_DEVICE_SIEMENS16 : DeviceInfoArray[i].num_ports = 16; break;

        //case PCI_DEVICE_RPLUS4 : DeviceInfoArray[i].num_ports = 4; break;
        //case PCI_DEVICE_RPLUS8 : DeviceInfoArray[i].num_ports = 8; break;

        default: DeviceInfoArray[i].num_ports = 32; break;
      }
      // printf("PCI,io:%x Ports:%d", boardio, DeviceInfoArray[i].num_ports);
    }
  }
  return 0;
}

/*----------------------------------------------------------------------
| GetDevInfo - Given Bus, Dev, Func find out the allocated irq and io-
|   address from the PCI config space.
|--------------------------------------------------------------------*/
static int GetDevInfo(DEVINFO *pci_config)
{
 int Bus, Dev, Func;
 // int result = 0;
 static WORD data_word;
 static BYTE data_byte;

  Bus  = pci_config->Bus_num;
  Dev  = pci_config->Dev_num;
  Func = pci_config->Fctn_num;
    
  if (!pci_read_word(Bus, Dev, Func, OFF_BASE_ADDR, (WORD far *) &data_word))
      pci_config->Base_Addr = data_word-1;
  else
  {
    return 1;
   // printf("ERROR reading config word \n\n");
  }

  if(!pci_read_byte(Bus, Dev, Func, OFF_INT_LINE, (BYTE far *) &data_byte))
     pci_config->Irq = data_byte;
  else
  {
    return 1;
   // printf("ERROR reading config byte \n\n");
  }

  if(!pci_read_word(Bus, Dev, Func, OFF_DEVICE_ID, (WORD far *) &data_word))
     pci_config->id = data_word;
  else
  {
    return 1;
   // printf("ERROR reading config byte \n\n");
  }

  return 0;
}

/*-----------------------------------------------------------------------
 FindDevs - This function finds all the PCI devices with the specified
  vendor &device id.
|------------------------------------------------------------------------*/
static int FindDevs(int vendor_id,
                    int max_device_id,
                    int *num_pci_boards,
                    DEVINFO *pci_config)
{
  int found = 0;
  int stat;
  int ctlindex;
  int devindex;
  static int bus_number;
  static int function_number;
  static int device_number;

     // loop thru all PCI slots
  for (ctlindex=0; ctlindex<MAX_PCI_BRDS; ctlindex++)
  {
    for (devindex=max_device_id; devindex>0; devindex--)
    {
      stat = pci_find_device(devindex,    // device ID
                             vendor_id,   // vendor ID
                             ctlindex,    // next index
                             (int far *) &bus_number,
                             (int far *) &function_number,
                             (int far *) &device_number);
      if (stat == 0) // ok
      {        /* found a device */
        //  if (verbose) printf("Found Board\n");
        if (*num_pci_boards >= MAX_PCI_BRDS)
          return 0;  // ok, enough PCI boards already

        pci_config[*num_pci_boards].DevId = devindex;
        pci_config[*num_pci_boards].Bus_num = bus_number;
        pci_config[*num_pci_boards].Dev_num = device_number;
        pci_config[*num_pci_boards].Fctn_num = function_number;
        if (*num_pci_boards < MAX_PCI_BRDS)
          (*num_pci_boards)++;
        else
          return 1;  // err: too many boards found.
        found = 1;
      }       /* found a device */
    }           /* device id loop */
  }       /* controller for loop */
  
  if (!found)
    return 2; // err: none found;

  return 0; // ok
}

/*-----------------------------------------------------------------------
| aaAutoConfig - Handle auto-configuration of the RocketPort boards.
|------------------------------------------------------------------------*/
int aaAutoConfig(int allow_pci, int irq)
{
 int stat,i;
 static DEVINFO brds[4];
 static int num_brds;
 WORD base_address, io;
 BYTE empty_val;
 int isa_brds;

  base_address = 0;

   for (i=0; i<4; i++)
   {
     brds[i].Base_Addr = 0;
     brds[i].Irq = 0;
     brds[i].id = 0xffff;
     brds[i].num_ports = 0;
   }

   if (allow_pci)
     aaFindPCIBoards(&brds[0], &num_brds);  // check for any PCI boards.

   if (num_brds != 0)
   {
#ifdef DEBUG_CONFIG
     printf("%d PCI boards found\n", num_brds);
     for (i=0; i<num_brds; i++)
     {
       printf("PCI board %d has addr:%xH, IRQ:%d Ports:%d\n", i+1,
         brds[i].Base_Addr,
         brds[i].Irq,
         brds[i].num_ports);
     }
#endif
   }

   if (num_brds == 4)
     base_address = 0;  // limit to 4 boards
   else if (scan_for_mudback(0x1c0) == 0)
     base_address = 0x180;
   else if (scan_for_mudback(0x180) == 0)
     base_address = 0x140;
   else if (scan_for_mudback(0x140) == 0)
     base_address = 0x100;
   else if (scan_for_mudback(0x200) == 0)
     base_address = 0x1C0;
   else if (num_brds == 0)
     return 1; // error, no boards found

   
   if (base_address != 0)
   {
     io = base_address + 0x40;
     sOutB(io+MUD_AIOPIC_EN, 0);  // turn off aiops
     sOutB(io+MUD_AIOP_ADDR, 0);  // turn off addressing
     empty_val = scan_40hex_space(base_address);
     if ((empty_val != 0) && (empty_val != 0xff))
     {
#ifdef DEBUG_CONFIG
       printf("Warning: Io-space in use!\n");
#endif
     }
     // sOutB(MudbacIO,(BYTE)(base_address >> 6));  /* set up AIOP I/O in MUDBAC */

     brds[num_brds].Base_Addr = base_address;
     brds[num_brds].id = 0x1000;  // mark as an ISA board
#ifdef DEBUG_CONFIG
     printf("ISA board has addr:%xH\n", brds[num_brds].Base_Addr);
#endif
     ++num_brds;
     isa_brds = 1;
     for (i=num_brds; i<4; i++)
     {
       // see if next card is present
       io += 0x400;
       stat = scan_for_mudback(io);
       // bug: stat = scan_for_mudback(base_address+(0x440*isa_brds));
       if (stat != 0)  // RocketPort ISA board not found
         break;

       // instead of using new 40H address space blocks, we are
       // going to alias 0x1000 up further.  This works in most machines,
       // the alternative is to allocate additional 40-hex io-blocks.

       brds[num_brds].Base_Addr = (isa_brds * 0x1000) + base_address;
       brds[num_brds].id = 0x1000;  // mark as an ISA board
       ++isa_brds;
       ++num_brds;
#ifdef DEBUG_CONFIG
       printf("ISA board has addr:%xH\n", brds[num_brds].Base_Addr);
#endif
     }
   }

   if (brds[0].id != 0x1000) // its a PCI board
   {
     if (irq != 1)  // set to 1 to use PCI selected interrupt
       brds[0].Irq = 0;
   }
   else
    brds[0].Irq = irq;

#ifdef DEBUG_CONFIG
   printf("Setting up %d boards\n", num_brds);
   for (i=0; i<4; i++)
   {
     printf("Brd %d at %xH\n", i, brds[i].Base_Addr);
   }
#endif

   stat =  aaSetConfig(brds[0].Irq,
                       brds[0].Base_Addr, brds[0].num_ports,
                       brds[1].Base_Addr, brds[1].num_ports,
                       brds[2].Base_Addr, brds[2].num_ports,
                       brds[3].Base_Addr, brds[3].num_ports);
#ifdef DEBUG_CONFIG
   if (stat)
     printf("Error %d from aaSetConfig.\n", stat);
#endif
   return 0;
}

/*-----------------------------------------------------------------------
| scan_40hex_space - scan 40H bytes of io-address space, since we have
    our hardware turned off, we should see an open bus, normally all 0xff's.
    We will allow all 0's, you never know with PC hardware.
|------------------------------------------------------------------------*/
int scan_40hex_space(WORD io_addr)
{
  BYTE val;
  int i;

  val = sInB(io_addr);
  if ((val != 0) && (val != 0xff))
  {
    return 1;  // return an error.
  }
  for (i=1; i<0x40; i++)
  {
    if (val != sInB(io_addr+i))
      return 2;  // error
  }
  return val;  // return the common(open bus) value, normally 0xff.
}

/*-----------------------------------------------------------------------
| scan_for_mudback - check the io-addr for a mudback chip.  Return zero
   if we found one, otherwise non-zero indicating what failed.  The mudback
   chip is the manager for ISA & MCA bus systems.  Default settings of
   180H are really setting mudback chip up at 1C0H, additional boards
   are at +400H, so 2nd-5C0H, 3rd-9C0H, 4th-DC0H.
|------------------------------------------------------------------------*/
int scan_for_mudback(WORD io_addr)
{
// int stat;
 BYTE old;

  //if (verbose > 1) printf("Checking Address:%xH\n",io_addr);
  sInB(io_addr+MUD_INT_STAT);   // clear out any low intr-status bits
  if (sInB(io_addr+MUD_INT_STAT) != 0xf0) // verify no intr-bits
    return 1;
  //if (verbose > 1) printf("Found Mudback Intr-Status = 0F0H\n");
  if ((sInB(io_addr+MUD_INT_EN) & INT_STROB) != 0) // int_strob always 0
    return 2;
  //if (verbose > 1) printf("Found INT_STROBE = 0H\n");

      // probably found one, now start writing to registers to check
                             // check operation of inter. enable reg.
  old = sInB(io_addr+MUD_INT_EN);
  sOutB(io_addr+MUD_INT_EN, 0);      // turn all bits off
  if (sInB(io_addr+MUD_INT_EN) != 0) // verify all bits read off
  {
    sOutB(io_addr+MUD_INT_EN, old);
    //if (verbose > 1) printf("Bad Retention Int. En Register <> 0H\n");
    return 3;  // err
  }

  sOutB(io_addr+MUD_INT_EN, 0xf7);  // set some bits
  if (sInB(io_addr+MUD_INT_EN) != 0xf3)  // check these bits(INT_STROBE won't set)
  {
    sOutB(io_addr+MUD_INT_EN, old);
    //if (verbose > 1) printf("Bad Retention Int. En Register <> 0F3H\n");
    return 4;  // err
  }
  sOutB(io_addr+MUD_INT_EN, 0);
  //if (verbose > 1) printf("Passed Retention Int. En Register Test\n");

                             // check operation of aiopic. enable reg.
  old = sInB(io_addr+MUD_AIOPIC_EN);
  sOutB(io_addr+MUD_AIOPIC_EN, 0xff);
  if (sInB(io_addr+MUD_AIOPIC_EN) != 0xff)
  {
    sOutB(io_addr+MUD_AIOPIC_EN, old);
    //if (verbose > 1) printf("Bad AIOPIC Enable Register <> 0FFH\n");
    return 5;
  }

  sOutB(io_addr+MUD_AIOPIC_EN, 0);
  if (sInB(io_addr+MUD_AIOPIC_EN) != 0)
  {
    sOutB(io_addr+MUD_AIOPIC_EN, old);
    // if (verbose > 1) printf("Bad AIOPIC Enable Register <> 0H\n");
    return 6;
  }
  return 0;
}

/*-----------------------------------------------------------------------
Function: aaSetConfig
Purpose:  Alternative to controller configuration information file
  setup by orginal install program.  Allows the application to setup
  the irq and address for each RocketPort Board.
Call:     aaSetConfig(irq, io1, io2, io3, io4)
          int irq; IRQ to use, 0 if no IRQ should be used.
          int io1; I/O Address for 1st board, 0 if None, 1 if PCI.
          int ports1; max number of ports on board.
          int io2; I/O Address for 2nd board, 0 if None, 1 if PCI.
          int ports2; max number of ports on board.
          int io3; I/O Address for 3rd board, 0 if None, 1 if PCI.
          int ports3; max number of ports on board.
          int io4; I/O Address for 4th board, 0 if None, 1 if PCI.
          int ports4; max number of ports on board.
Return:   int: 0 OK, otherwise error code.
|-------------------------------------------------------------------------*/
int aaSetConfig(int irq, unsigned int io1, unsigned int ports1,
                         unsigned int io2, unsigned int ports2,
                         unsigned int io3, unsigned int ports3,
                         unsigned int io4, unsigned int ports4)
{
  int i,j;
  CF_T *CfgP;        /* ptr to structure holding config info for all ctls */
  WORD isa_mudback_io_addr = 0;

   /* Initialize config info struct to known values */
   CfgP = &aaCfg;
   for(i = 0;i < CTL_SIZE;i++)
   {
      CfgP->CfCtl[i].MudbacIO = 0;
      CfgP->CfCtl[i].MaxPorts = 0;
      CfgP->CfCtl[i].BusType  = BUS_ISA;

      for(j = 0;j < AIOP_CTL_SIZE;j++)         /* AIOP aliases */
      {
         CfgP->CfCtl[i].AiopIO[j] = 0;
      }
   }

   if (irq > 0)
     CfgP->Irq = irq;

   CfgP->NumCtl = 0;
   if (io1 > 0)
   {
     CfgP->CfCtl[0].AiopIO[0] = io1;
     CfgP->CfCtl[0].MaxPorts = ports1;
     CfgP->NumCtl = 1;
   }
   if (io2 > 0)
   {
     CfgP->NumCtl = 2;
     CfgP->CfCtl[1].AiopIO[0] = io2;
     CfgP->CfCtl[1].MaxPorts = ports2;
   }
   if (io3 > 0)
   {
     CfgP->NumCtl = 3;
     CfgP->CfCtl[2].AiopIO[0] = io3;
     CfgP->CfCtl[2].MaxPorts = ports3;
   }
   if (io4 > 0)
   {
     CfgP->NumCtl = 4;
     CfgP->CfCtl[3].AiopIO[0] = io4;
     CfgP->CfCtl[3].MaxPorts = ports4;
   }

   for(i = 0;i < CfgP->NumCtl;i++)
   {
      if (CfgP->CfCtl[i].MaxPorts == 0)
      {
        // for ISA, we allow entry of 0, then set to 32 here
        // and let init routine figure out how many are really
        // there
        CfgP->CfCtl[i].MaxPorts = 32;
      }

      if ((CfgP->CfCtl[i].AiopIO[0] > 0x3c0) && (isa_mudback_io_addr == 0))
      {
        //---- assume PCI
        CfgP->CfCtl[i].BusType  = BUS_PCI;

        for(j = 1;j < AIOP_CTL_SIZE;j++)
           CfgP->CfCtl[i].AiopIO[j] = CfgP->CfCtl[i].AiopIO[j - 1] + 0x40;
        //printf("pci ctrl:%d at %xH\n", i, CfgP->CfCtl[i].AiopIO[0]);
        CfgP->CfCtl[0].MudbacIO = CfgP->CfCtl[0].AiopIO[0];  // no mudback on pci board
      }
      else  // ISA
      {
        CfgP->CfCtl[i].BusType  = BUS_ISA;

        for(j = 1;j < AIOP_CTL_SIZE;j++)      /* AIOP aliases */
           CfgP->CfCtl[i].AiopIO[j] = CfgP->CfCtl[i].AiopIO[j - 1] + 0x400;

        if (isa_mudback_io_addr == 0)
          isa_mudback_io_addr = CfgP->CfCtl[i].AiopIO[0] + 0x40;
        else
          isa_mudback_io_addr += 0x400;  // alias up by 0x400.

        CfgP->CfCtl[i].MudbacIO = isa_mudback_io_addr;  // alias up by 0x400.

        //printf("isa ctrl:%d at %xH\n", i, CfgP->CfCtl[i].AiopIO[0]);
      }
   }

   config_setup = 1;  // don't try to read config file later
   return 0;  // ok
}

#ifdef MICROCHAN_NO_SUPPORT
/****************************************************************************
Function:  aaReadConfig
Purpose:   Read controller configuration information from POS registers
Call:      aaReadConfig()
Return:    int: Always returns TRUE to indicate POS registers read successfully.
Warnings:  This routine provides NO error checking.
           This is an API privaleged function.
*/
int aaReadConfig(void)
{
   int j;                /* counter */
   int Slot;             /* MCA slot counter */
   int Ctl;              /* controller number */
   int POSID;            /* Board ID */
   char POS1;            /* MUDBAC base IO address (MSByte) */
   char POS2;            /* AIOP base IO address (MSByte) */
   char POS3;            /* store IRQ number in lower 4 bits */
   CF_T *CfgP;        /* ptr to structure holding config info for all ctls */

   CfgP = &aaCfg;
   Slot = 1;
   Ctl = 0;
   while((Slot <= MAXMCASLOTS) && (Ctl < CTL_SIZE))
   {
      /* read slot's POS registers */
      sOutW(0x96,Slot+7);         /* enable slot POS register response */
      POSID = sInW(0x100);
      POS1 = sInB(0x103);
      POS2 = sInB(0x104);
      POS3 = sInB(0x105) & 0xf;   /* strip to 4 bits */
      sOutW(0x96,0);              /* disable slot POS register response */

      switch(POSID)                    /* check controller ID */
      {
         case MCA8CHAN:                /* controller found in this slot */
         case MCA16CHAN:
         case MCA32CHAN:
            CfgP->CfCtl[Ctl].MudbacIO = (WORD)POS1 << 8; /* MUDBAC I/O */
            CfgP->CfCtl[Ctl].AiopIO[0] = (WORD)POS2 << 8; /* 1st AIOP I/O */
            for(j = 1;j < AIOP_CTL_SIZE;j++) /* AIOP aliases */
               CfgP->CfCtl[Ctl].AiopIO[j] = CfgP->CfCtl[Ctl].AiopIO[j - 1] + 0x40;
            if(Ctl == 0)               /* IRQ on 1st controller only */
               CfgP->Irq = (int)POS3 & 0x0f;
            Ctl++;
            break;
         default:
            break;
      }
      Slot++;
   }
   CfgP->NumCtl = Ctl;
   return(TRUE);
}
#endif

/******************************************************************************
Function: aaReadConfig
Purpose:  Read the controller configuration information file.
Call:     aaReadConfig()
Return:   int: TRUE if file read successfully, else FALSE.
Comments: The configuration information file contains configuration
          information for all the controllers in a system.  This file
          is a simple text file where each line represents
          information for a single controller.  Also, each parameter
          must be seperated by commas and must be in the correct order
          as defined.
Warnings: The config data file can only be 512 bytes long!!
          This routine provides NO error checking.
          This is an API privaleged function.
*/
int aaReadConfig(void)
{
  static char filebuf[512];      /* buffer to load file into */
  int  fp;                       /* file pointer (handle) */
  int i;
  char *ConfigFile;              /* ptr to configuration file name */
  unsigned int brd_addr[4];
  unsigned int brd_ports[4];
  int irq;

  irq = 0;
  for (i=0; i<4; i++)
  {
    brd_addr[i] = 0;
    brd_ports[i] = 0;
  }

   /* Open the config file and read in the data */
   ConfigFile = getenv("ROCKETCFG"); /* get cfg file name from environment */

   /* Open the config file and read in the data */
   ConfigFile = getenv("ROCKETCFG"); /* get cfg file name from environment */
   if((fp = open(ConfigFile,O_RDONLY | O_BINARY)) < 0)
   {
      return(FALSE);
   }
   else
   {
      read(fp,filebuf,sizeof(filebuf));
   }

   close(fp);

   /* Convert config file info to IRQ and addresses */
   i = sscanf(filebuf,"%d %x%d %x%d %x%d %x%d",
              &irq,
              &brd_addr[0], &brd_ports[0],
              &brd_addr[1], &brd_ports[1],
              &brd_addr[2], &brd_ports[2],
              &brd_addr[3], &brd_ports[3]);

   if(i < 3)
      return(FALSE);

   for (i=0; i<4; i++)
   {
     if (brd_ports[i] > 32)
       return FALSE;
   }

   aaSetConfig(irq,
              brd_addr[0], brd_ports[0],
              brd_addr[1], brd_ports[1],
              brd_addr[2], brd_ports[2],
              brd_addr[3], brd_ports[3]);

   return(TRUE);
}

/*******************************************************************************
Function:  aaInit
Purpose:   Controller and API initialization
Call:      aaInit()
Return:    WORD:       NO_ERR         if no initialization errors
                         ERR_NOCFGFILE  if can't open configuration file
                         ERR_ALLOCDEV   if can't allocate device structure
                         ERR_CTLINIT    if controller initialization error
                         ERR_CHANINIT   if channel initialization error
                         ERR_DEVSIZE    if invalid number of devices found
                         ERR_CTLSIZE    if invalid number of controllers found
Comments:  This function must be called once before calling any other API
           function except aaInstallCtrlCHandler().  The controller
           initialization parameters will be obtained from the configuration
           file given by enviornment variable ROCKETCFG if this is an ISA
           system, or from the POS registers if this is a Microchannel system.
Warnings:  Once aaInit() has been called, aaExit() must be called before
           exiting the application program.  This is true regardless of
           the value returned by aaInit().

           If the application program can be exited using the CTRL+C or
           CTRL+BREAK keys, then the default DOS CTRL+C handler must be
           replaced with a handler that calls aaExit().  Function
           aaInstallCtrlCHandler() can be used for this purpose.
*/
WORD aaInit(void)
{
   int Ctl;                            /* controller number */
   static int PrevInit = FALSE;        /* aaInit() previously called */
   int Ch;                             /* channel number */
   int Dev;                            /* device number */
   CONTROLLER_T *CtlP;                 /* ptr to controller structure */
   DEV_T *DevP;                    /* ptr to device structure */
   int Aiop;                           /* AIOP number */
   unsigned long StartTime;            /* controller time */
   long i;
   unsigned char x;
   int IrqNum;
   int max_aiops;
#ifdef TRY_EVENT_IRQS
   int freq = FREQ_DIS;
   int periodic_only = FALSE;
#else
   int freq = FREQ_274HZ;
   int periodic_only = TRUE;
#endif

   if(PrevInit)                        /* only allow aaInit() to execute once */
      return(NO_ERR);
   aaIrqSave = 0;                      /* indicate no IRQ installed */
   PrevInit = TRUE;

   /* Read configuration information from configuration file */
   for(Ctl = 0;Ctl < CTL_SIZE;Ctl++)
   {
      aaCfg.CfCtl[Ctl].StartDev = -1;
      aaCfg.CfCtl[Ctl].NumChan = 0;
   }

   if (!config_setup)
   {
     if(!aaReadConfig())                 /* get configuration information */
     {
       return(ERR_NOCFGFILE);           /* can't read configuration file */
     }
   }

   /* Allocate device structures & initialize to default values */
   aaCtlSize = aaCfg.NumCtl;
   if((aaCtlSize < 1) || (aaCtlSize > CTL_SIZE))
      return(ERR_CTLSIZE);

   /* Initialize controller and channel structures */
   Dev = 0;
   for(Ctl = 0;Ctl < aaCtlSize;Ctl++)
   {
      CtlP = &sController[Ctl];
      if(Ctl == 0)                     /* periodic int on 1st ctl only */
         IrqNum = aaCfg.Irq;
      else
         IrqNum = 0;

      CtlP->BusType = aaCfg.CfCtl[Ctl].BusType;
      if (aaCfg.CfCtl[Ctl].AiopIO[0] < 0x400)
      {
        // if ISA board, then init call can figure out how many ports.
        aaCfg.CfCtl[Ctl].MaxPorts = 32;
      }

      if (aaCfg.CfCtl[Ctl].MaxPorts > 16)
        max_aiops = 4;
      else if (aaCfg.CfCtl[Ctl].MaxPorts > 8)
        max_aiops = 2;
      else if (aaCfg.CfCtl[Ctl].MaxPorts > 0)
        max_aiops = 1;

      if(sInitController(CtlP,
                         Ctl,
                         aaCfg.CfCtl[Ctl].MudbacIO,
                         aaCfg.CfCtl[Ctl].AiopIO,
                         max_aiops,
                         IrqNum,
                         (BYTE) freq,
                         periodic_only) == CTLID_NULL)
      {
         return(ERR_CTLINIT + Ctl);    /* controller initialization error */
      }
      TEST_puts("ptA");
      aaCfg.CfCtl[Ctl].StartDev = Dev;
      for(Aiop = 0;Aiop < sGetControllerNumAiop(CtlP);Aiop++)
      {
         TEST_puts("ptB");
         sEnAiop(CtlP,Aiop);
         for(Ch=0; Ch < CtlP->AiopNumChan[Aiop]; Ch++)
         {
           TEST_puts("ptC");
            if((DevP = (DEV_T *)malloc(sizeof(DEV_T))) == NULL)
               return(ERR_ALLOCDEV);         /* can't allocate a device structure */

            aaDevices[Dev] = DevP;
            DevP->Dev = Dev;
            DevP->Hnd = 0;
            DevP->Status = 0;
            DevP->BreakTime = 0;

            // if irq's not used we cannot assume we have a tick-counter
            // to work with.
            if (IrqNum != 0)
                 DevP->CloseDelay = (CLOSE_TBEDLY * 1000)/(TICK_PERIOD/1000);
            else DevP->CloseDelay = 0;

            DevP->Chan.CtlP = NULL;
            DevP->Chan.AiopNum = -1;
            DevP->Chan.ChanNum = -1;

            sSetDevMap(Ctl,Aiop,Ch,Dev);
            if(sInitChan(CtlP,&DevP->Chan,Aiop,Ch))
            {
               return(ERR_CHANINIT + Dev); /* channel initialization error */
            }
            Dev++;
         }
         aaCfg.CfCtl[Ctl].NumChan += Ch;
      }
   }
   TEST_puts("ptD");
   aaDevSize = Dev;
   if((aaDevSize < 1) || (aaDevSize > DEV_SIZE))
      return(ERR_DEVSIZE);

   TEST_puts("ptE");
   /* Connect to an IRQ */
   aaControllerTime = 0;               /* initialize controller time */
   aaInstallIrq(aaCfg.Irq,aaIntHandler);

   TEST_puts("ptF");
   /* Install the default event functions */
   aaInstallMdmChgEvent(aaMdmChgEventNull);
   aaInstallRxEvent(aaRxEventNull);
   aaInstallPeriodicEvent(aaPeriodicEventNull);

   if (aaCfg.Irq == 0)  // no interrupts used
     return(NO_ERR);

   /* Tell the first controller to start sending periodic interrupts */
   CtlP = &sController[0];
   sEnGlobalInt(CtlP);

   aaEnableInt();                      /* enable interrupts */

   TEST_puts("ptF");
   /* Verify interrupts are occuring */
   StartTime = aaControllerTime;
   for(i = 0;i < 750000L;i++)    /* wait up to .05 seconds */
   {
      /* This MUDBAC access each provides minimum 200nS delay */
      x = sGetControllerIntStatus(CtlP);
      x = x;

      /* If controller time has changed interrupts are occuring */
      if(aaControllerTime != StartTime)
      {
         TEST_puts("ptG");
         return(NO_ERR);
      }
   }
   TEST_puts("ptH");

//   free(DevP);

   /* If we get here the IRQ's ISR has been installed but interrupts are
      not occuring.  Return an error */
   return(ERR_NOINT);
} /* end aaInit */

/*******************************************************************************
Function:  aaInstallCtrlCHandler
Purpose:   Install a handler for the CTRL+C key interrupt.
Call:      aaInstallCtrlCHandler()
Return:    void
Comments:  This function replaces the current CTRL+C (interrupt 23H) handler
           with a handler that performs the following actions:
           1. Calls aaExit().
           2. Sets the carry flag to signal DOS to terminate the application.
           3. Executes a far return.

           DOS will restore the original CTRL+C handler when terminating
           the application.

           aaInstallCtrlCHandler() is the only API function that may be called
           before calling aaInit().  If aaInstallCtrlCHandler() is to be used,
           it is recommended that it be called either immediately before or
           immediately after the call to aaInit().

           If you want different CTRL+C processing, you must write and install
           your own custom CTRL+C handler.  Refer to the "Microsoft MS-DOS
           Programmer's Reference" for more information.  To aid in writing
           your own handler the source code for aaInstallCtrlCHandler() and
           the handler it installs are given below:

           void aaInstallCtrlCHandler(void)
           {
              _dos_setvect(0x23,(void (interrupt far *)())aaCtrlCIntHandler);
           }

           void far aaCtrlCIntHandler(void)
           {
              aaExit();
              asm stc;
           }

Warnings:  If the application program can be exited using the CTRL+C or
           CTRL+BREAK keys, then the default DOS CTRL+C handler must be
           replaced with a handler that calls aaExit().  Function
           aaInstallCtrlCHandler() can be used for this purpose.
*/
void aaInstallCtrlCHandler(void)
{
   _dos_setvect(CTRLC_INT,(void (interrupt far *)())aaCtrlCIntHandler);
}

/*******************************************************************************
Function: aaGetCtlStatus
Purpose:  Get controller status, including the first device number on the
          controller and the number of devices on the controller.
Call:     aaGetCtlStatus(CtlNum,FirstDevP,NumDevP)
          int CtlNum; Controller number to get status on.
          int *FirstDevP; Pointer to variable where first device number
                          on this controller will be returned.
          int *NumDevP; Pointer to variable where number of devices on
                        this controller will be returned
Return:   int: NO_ERR:    Controller is installed
               ERR_NOCTL: Controller is not installed
Comments: The CtlNum parameter identifies which RocketPort controller to get
          the status of.  Controllers are numbered sequentially beginning with
          0.  For ISA systems controller 0 will be the first controller whose
          address appears in the configuration file given by enviornment
          variable ROCKETCFG.  For Microchannel systems controller 0 will be
          the controller located in the lowest numbered slot in the system.

          The contents of the FirstDevP and NumDevP parameters are modified
          only if NO_ERR is returned.
*/
int aaGetCtlStatus(int CtlNum,int *FirstDevP,int *NumDevP)
{
   if(CtlNum < 0 || CtlNum >= aaCtlSize)
      return(ERR_NOCTL);
   if(aaCfg.CfCtl[CtlNum].StartDev == -1)
      return(ERR_NOCTL);

   *FirstDevP = aaCfg.CfCtl[CtlNum].StartDev;
   *NumDevP = aaCfg.CfCtl[CtlNum].NumChan;
   return(NO_ERR);
}

/******************************************************************************
Function: aaOpen
Purpose:  Open a device for reading and/or writing.
Call:     aaOpen(Dev,OpenType,Baud,Parity,DataBits,StopBits,FlowCtl,
                 DetectEn,ModemCtl);
          int Dev; Device Number
          WORD OpenType; COM_TX for transmit and/or COM_RX for receive.
          unsigned long Baud; One of the COM_BAUD_XX flags or actual baud.
          unsigned char Parity; One of: COM_PAR_NONE, COM_PAR_EVEN, COM_PAR_ODD.
          unsigned char DataBits: One of COM_DATABIT_7, COM_DATABIT_8.
          unsigned char StopBits: One of COM_STOPBIT_1, COM_STOPBIT_2.
          WORD FlowCtl: Flow control flag, can be COM_FLOW_NONE or any
                          combination of: COM_FLOW_IS, COM_FLOW_OS,
                          COM_FLOW_IH, COM_FLOW_OH, COM_FLOW_OXANY.
          WORD DetectEn: Detection enable flags, can be any combination
             of the following:
                COM_DEN_NONE:    No error detection enabled
                COM_DEN_MDM:     Enable modem input (DSR, CD, or CTS) change
                                 detection.
                COM_DEN_RDA:     Enable Rx data available detection.
          unsigned char ModemCtl: Modem control lines to turn ON, can be
                            COM_MDM_RTS and/or COM_MDM_DTR.  If the flag is not
                            set the line will be OFF.  If hardware flow control
                            is in use for a modem line, it's flag has no
                            effect.
Return:   int: NO_ERR           if successful
               ERR_DEV          if device number out of range
               ERR_OPENTYPE     if invalid open type flag
               ERR_BAUDRATE     if invalid baud rate flag
               ERR_PAR          if invalid parity bits flag
               ERR_DATAB        if invalid data bits flag
               ERR_STOPB        if invalid stop bits flag
               ERR_FLOW         if invalid flow control bits flag
               ERR_DETECT       if invalid detect enable flag
               ERR_MDMCTL       if invalid modem control flag
               ERR_ALREADYOPEN  if device already open error flag
Comments: If this device has been opened previously and is still open, this
          function will fail and return an ERR_ALREADYOPEN error.
Warnings: This function disables and enables interrupts.
*/
int aaOpen(int Dev,WORD OpenType,unsigned long Baud,
                 unsigned char Parity,unsigned char DataBits,
                 unsigned char StopBits, WORD FlowCtl,
                 WORD DetectEn, unsigned char ModemCtl)
{
   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */
   WORD Status;                      /* device status */
   WORD IntFlags;                    /* interrupt enable flags */
   unsigned char RxTrig;               /* Rx FIFO trigger level */
   unsigned char x;

   /* Error check each parameter and fail if error found */
   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   if(aaCheckOpenType(OpenType) < 0)   /* check Open Type */
      return(ERR_OPENTYPE);
   if(aaCheckBaudRate(Baud) < 0)       /* check Baud Rate */
      return(ERR_BAUDRATE);
   if(aaCheckParity(Parity) < 0)       /* check Parity */
      return(ERR_PAR);
   if(aaCheckDataB(DataBits) < 0)      /* check Data Bits */
      return(ERR_DATAB);
   if(aaCheckStopB(StopBits) < 0)      /* check Stop Bits */
      return(ERR_STOPB);
   if((FlowCtl & ~(FLOWCHECK)) != 0)   /* check Flow Control */
      return(ERR_FLOW);
   if((DetectEn & ~(DETECTCHECK)) != 0)  /* check Detect Enable */
      return(ERR_DETECT);
   if((ModemCtl & ~(MDMCTLCHECK)) != 0)  /* check Modem Control */
      return(ERR_MDMCTL);
   DevP = aaDevices[Dev];
   ChP = &DevP->Chan;
   if(DevP->Status != 0)
      return(ERR_ALREADYOPEN);

   /* Determine opening device status */
   Status = OpenType | COM_OPEN;
   Status = Status | FlowCtl | COM_TXFLOW_ON | COM_RXFLOW_ON;

   /* Determine interrupt enables */
   IntFlags = 0;
   RxTrig = TRIG_NO;
   if(OpenType & COM_RX)
   {
      if(DetectEn & COM_DEN_RDA)
         IntFlags |= RXINT_EN;         /* int on Rx FIFO at trigger level */
      RxTrig = TRIG_1;                 /* trigger level 1 character */
   }
   if(DetectEn & COM_DEN_MDM)
      IntFlags |= MCINT_EN;            /* int on modem control change */
   if(IntFlags)
      IntFlags |= CHANINT_EN;          /* there are channel interrupts enabled */

   if(FlowCtl & COM_FLOW_IH)           /* if input hardware flow don't turn RTS on */
      ModemCtl &= ~COM_MDM_RTS;
   if(FlowCtl & COM_FLOW_IS)           /* input sw flow ctl must Tx XON/XOFF */
      Status |= COM_TX;
   if(FlowCtl & COM_FLOW_OS)           /* output sw flow ctl must Rx XON/XOFF */
      Status |= COM_RX;

   /* Initialize the channel */
   aaDisableInt();                     /* disable interrupts */
   sDisInterrupts(ChP,TXINT_EN | RXINT_EN | SRCINT_EN | MCINT_EN | CHANINT_EN);
   sDisTransmit(ChP);
   sDisRxStatusMode(ChP);
   sFlushRxFIFO(ChP);
   sFlushTxFIFO(ChP);
   sDisRTSFlowCtl(ChP);
   sDisCTSFlowCtl(ChP);
   //sDisRxSoftFlowCtl(ChP);
   sDisTxSoftFlowCtl(ChP);
   sDisIXANY(ChP);
   sClrTxXOFF(ChP);

   aaSetBaud(Dev, Baud);

   switch(Parity)                      /* set parity */
   {
      case COM_PAR_NONE:
         sDisParity(ChP);
         break;
      case COM_PAR_EVEN:
         sEnParity(ChP);
         sSetEvenParity(ChP);
         break;
      case COM_PAR_ODD:
         sEnParity(ChP);
         sSetOddParity(ChP);
         break;
   };
   if(DataBits == COM_DATABIT_7)       /* set data bits */
   {
      sSetData7(ChP);
      sSetRxMask(ChP,0x7f);
   }
   else
   {
      sSetData8(ChP);
      sSetRxMask(ChP,0xff);
   }
   if(StopBits == COM_STOPBIT_1)       /* set stop bits */
   {
      sSetStop1(ChP);
   }
   else
   {
      sSetStop2(ChP);
   }
   if(ModemCtl & COM_MDM_RTS)          /* turn on/off modem control lines */
   {
      sSetRTS(ChP);
   }
   else
   {
      sClrRTS(ChP);
   }
   if(ModemCtl & COM_MDM_DTR)
   {
      sSetDTR(ChP);
   }
   else
   {
      sClrDTR(ChP);
   }
#ifdef INTERNAL_ISFLOW
   if(FlowCtl & COM_FLOW_IS)           /* flow control */
   {
      sEnRxSoftFlowCtl(ChP);
      //sSetTxSize(ChP,MAXTX_SIZE / 2 - 8);
   }
   else
   {
      //sDisRxSoftFlowCtl(ChP);
      //sSetTxSize(ChP,MAXTX_SIZE);
   }
#endif
   if(FlowCtl & COM_FLOW_OS)           /* flow control */
   {
      sEnTxSoftFlowCtl(ChP);
   }
   if(FlowCtl & COM_FLOW_OXANY)
   {
      sEnIXANY(ChP);
   }
   if(FlowCtl & COM_FLOW_IH)
   {
      sEnRTSFlowCtl(ChP);
   }
   if(FlowCtl & COM_FLOW_OH)
   {
      sEnCTSFlowCtl(ChP);
   }
   //sSetRxXOFFChar(ChP,0x13);
   //sSetRxXONChar(ChP,0x11);
   if(Status & COM_RX)
   {
      sEnRxFIFO(ChP);                     /* enable Rx FIFO */
      sSetRxTrigger(ChP,RxTrig);          /* Rx FIFO trigger level */
      sStartRxProcessor(ChP);
   }
   else
   {
      sStopRxProcessor(ChP);
   }
   if(Status & COM_TX)
   {
      sEnTransmit(ChP);                /* enable tranamit */
   }
   if(IntFlags)                        /* enable channel interrupts */
   {
      sEnInterrupts(ChP,IntFlags);
   }
   x = sGetChanIntID(ChP); x = x;      /* clear any modem input changes */
   DevP->DetectEn = DetectEn;
   DevP->Status = Status;
   aaEnableInt();                      /* enable interrupts */
   return(NO_ERR);
}  /* end aaOpen */

/******************************************************************************
Function: aaReconfigure
Purpose:  Reconfigure a device's communication parameters.
Call:     aaReconfigure(Dev,Baud,Parity,DataBits,StopBits,FlowCtl,DetectEn);
          int dev; Device Number
          unsigned char Baud; One of the baud rate flags or actual baud rate.
          unsigned char Parity; One of: COM_PAR_NONE, COM_PAR_EVEN, COM_PAR_ODD.
          unsigned char DataBits: One of COM_DATABIT_7, COM_DATABIT_8.
          unsigned char StopBits: One of COM_STOPBIT_1, COM_STOPBIT_2.
          WORD FlowCtl: Flow control flag, can be COM_FLOW_NONE or any
                          combination of: COM_FLOW_IS, COM_FLOW_OS,
                          COM_FLOW_IH, COM_FLOW_OH, COM_FLOW_OXANY.
          WORD DetectEn: Detection enable flags, can be any combination
             of the following:
                COM_DEN_NONE:    No error detection enabled
                COM_DEN_MDM:     Enable modem input (DSR, CD, or CTS) change
                                 detection
                COM_DEN_RDA:     Enable Rx data available detection
Return:   int: NO_ERR            if successful
               ERR_DEV           if device number out of range
               ERR_BAUDRATE      if invalid baud rate flag
               ERR_PAR           if invalid parity bits flag
               ERR_DATAB         if invalid data bits flag
               ERR_STOPB         if invalid stop bits flag
               ERR_FLOW          if invalid flow control bits flag
               ERR_DETECT        if invalid detect enable flag
               ERR_NOTOPEN       if device not open error flag
Warnings: This function disables and enables interrupts.
*/
int aaReconfigure(int Dev,unsigned long Baud,unsigned char Parity,
                        unsigned char DataBits,unsigned char StopBits,
                        WORD FlowCtl,WORD DetectEn)
{
   WORD IntFlags;                    /* interrupt enable flags */
   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */

   /* Error check each parameter and fail if error found */
   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   if(aaCheckBaudRate(Baud) < 0)       /* check Baud Rate */
      return(ERR_BAUDRATE);
   if(aaCheckParity(Parity) < 0)       /* check Parity */
      return(ERR_PAR);
   if(aaCheckDataB(DataBits) < 0)      /* check Data Bits */
      return(ERR_DATAB);
   if(aaCheckStopB(StopBits) < 0)      /* check Stop Bits */
      return(ERR_STOPB);
   if((FlowCtl & ~(FLOWCHECK)) != 0)   /* check Flow Control */
      return(ERR_FLOW);
   if((DetectEn & ~(DETECTCHECK)) != 0)  /* check Detect Enable */
      return(ERR_DETECT);
   DevP = aaDevices[Dev];
   ChP = &DevP->Chan;
   if((DevP->Status & COM_OPEN) == 0)  /* check device open */
      return(ERR_NOTOPEN);

   /* Reconfigure the channel */
   aaDisableInt();                     /* disable interrupts */
   sDisInterrupts(ChP,TXINT_EN | RXINT_EN | SRCINT_EN | MCINT_EN | CHANINT_EN);
   aaSetBaud(Dev, Baud);
   switch(Parity)                      /* set parity */
   {
      case COM_PAR_NONE:
         sDisParity(ChP);
         break;
      case COM_PAR_EVEN:
         sEnParity(ChP);
         sSetEvenParity(ChP);
         break;
      case COM_PAR_ODD:
         sEnParity(ChP);
         sSetOddParity(ChP);
         break;
   };
   if(DataBits == COM_DATABIT_7)       /* set data bits */
   {
      sSetData7(ChP);
      sSetRxMask(ChP,0x7f);
   }
   else
   {
      sSetData8(ChP);
      sSetRxMask(ChP,0xff);
   }
   if(StopBits == COM_STOPBIT_1)       /* set stop bits */
   {
      sSetStop1(ChP);
   }
   else
   {
      sSetStop2(ChP);
   }
   if(FlowCtl & COM_FLOW_IS)           /* flow control */
   {
#ifdef INTERNAL_ISFLOW
      sEnRxSoftFlowCtl(ChP);
      //sSetTxSize(ChP,MAXTX_SIZE / 2 - 8);
#endif
      DevP->Status |= COM_FLOW_IS;
   }
   else
   {
#ifdef INTERNAL_ISFLOW
      //sDisRxSoftFlowCtl(ChP);
      //sSetTxSize(ChP,MAXTX_SIZE);
#endif
      DevP->Status &= ~COM_FLOW_IS;
   }
   if(FlowCtl & COM_FLOW_OS)
   {
      sEnTxSoftFlowCtl(ChP);
      DevP->Status |= COM_FLOW_OS;
   }
   else
   {
      sDisTxSoftFlowCtl(ChP);
      DevP->Status &= ~COM_FLOW_OS;
   }
   if(FlowCtl & COM_FLOW_OXANY)
   {
      sEnIXANY(ChP);
      DevP->Status |= COM_FLOW_OXANY;
   }
   else
   {
      sDisIXANY(ChP);
      DevP->Status &= ~COM_FLOW_OXANY;
   }
   if(FlowCtl & COM_FLOW_IH)
   {
      sEnRTSFlowCtl(ChP);
      DevP->Status |= COM_FLOW_IH;
   }
   else
   {
      sDisRTSFlowCtl(ChP);
      DevP->Status &= ~COM_FLOW_IH;
   }
   if(FlowCtl & COM_FLOW_OH)
   {
      sEnCTSFlowCtl(ChP);
      DevP->Status |= COM_FLOW_OH;
   }
   else
   {
      sDisCTSFlowCtl(ChP);
      DevP->Status &= ~COM_FLOW_OH;
   }
   DevP->Status |= (COM_TXFLOW_ON | COM_RXFLOW_ON);

   IntFlags = 0;                       /* set interrupts and detect enables */
   if(DevP->Status & COM_RX)
   {
      if(DetectEn & COM_DEN_RDA)
         IntFlags |= RXINT_EN;         /* int on Rx FIFO at trigger level */
   }
   if(DetectEn & COM_DEN_MDM)
   {
      IntFlags |= MCINT_EN;            /* int on modem control change */
   }
   if(IntFlags)
   {
      IntFlags |= CHANINT_EN;          /* there are channel interrupts enabled */
      sEnInterrupts(ChP,IntFlags);
   }
   DevP->DetectEn = DetectEn;
   aaEnableInt();                     /* reenable interrupts */
   return(NO_ERR);
}  /* end aaReconfigure */

/*-----------------------------------------------------------------------------
Function: aaSetBaud
Purpose:  Setup a baud rate, handles non-standard baud rates.
Call:     aaSetBaud(Dev, BaudRate);
Return:   Returns NO_ERR on success, returns ERR_BAUDRATE, ERR_DEV if can't do.
|---------------------------------------------------------------------------*/
int aaSetBaud(int Dev, unsigned long BaudRate)
{
  DEV_T *DevP;                    /* ptr to device structure */
  CHANPTR_T ChP;                      /* ptr to channel structure */
  long newbaud;

    // if using baud-index(literals) or actual baud
   if (BaudRate < MAXNUMBAUDS)
         newbaud = aaBaud[(int)BaudRate];
   else  newbaud = BaudRate;

   /* Error Check each parameter and fail if error found */
   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   DevP = aaDevices[Dev];

   ChP = &DevP->Chan;

   // first just validate that we can do the baud rate with our hardware
   if (sSetBaudRate(ChP, newbaud, 0) != 0)  // error, can't do baud
     return ERR_BAUDRATE;

   sSetBaudRate(ChP, newbaud, 1);  // actually set it.
  return NO_ERR;
}

/******************************************************************************
Function: aaClose
Purpose:  Close a device.
Call:     aaClose(Dev,ModemCtl)
          int Dev: Device number
          unsigned char ModemCtl: Modem control lines to turn OFF, can be
                                COM_MDM_RTS or COM_MDM_DTR.  If the flag is
                                not set the state of the modem line will not
                                be changed.
Return:   int: NO_ERR if successful
               ERR_DEV if device number out of range
               ERR_MDMCTL if invalid modem control flag
               ERR_NOTOPEN if device not open
Comments: This function will wait for the device's transmit buffer to drain
          before completing the close.  The maximum wait time defaults
          to CLOSE_TBEDLY, but can be changed with aaSetCloseDelay().
Warnings: This function disables and enables interrupts.
*/
int aaClose(int Dev,unsigned char ModemCtl)
{
   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */
   int TxCnt;                          /* Tx data left on this device */
   long DelayCnt;                      /* delay tick count waiting for TBE */

   /* Error Check each parameter and fail if error found */
   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   if((ModemCtl & ~(MDMCTLCHECK)) != 0) /* check modem control */
      return(ERR_MDMCTL);
   DevP = aaDevices[Dev];
   ChP = &DevP->Chan;
   if((DevP->Status & COM_OPEN) == 0)  /* if not open, don't close it */
      return(ERR_NOTOPEN);

   /* Wait for transmit FIFO to drain */
   DelayCnt = 0;
   TxCnt = sGetTxCnt(ChP);
   while((TxCnt > 0) && (DelayCnt < DevP->CloseDelay))
   {
      /* Normally would expect aaDelay(1) to result in a 1 tick
         delay.  However, if many aaDelay() calls are made rapidly in
         succession each call aaDelay(1) actually results in two ticks
         because aaDelay() waits for controller time to roll over before
         beginning to count down the delay interval. */
      aaDelay(1);                      /* wait 2 ticks */
      TxCnt = sGetTxCnt(ChP);
      DelayCnt += 2;
   }

   /* Clear device structure members */
   aaDisableInt();                     /* disable interrupts */
   DevP->Status = 0;
   DevP->DetectEn = 0;
   DevP->Hnd = 0;

   /* Close the channel */
   sDisInterrupts(ChP,TXINT_EN | RXINT_EN | SRCINT_EN | MCINT_EN | CHANINT_EN);
   sDisTransmit(ChP);
   //sDisRxSoftFlowCtl(ChP);
   sDisTxSoftFlowCtl(ChP);
   sDisRTSFlowCtl(ChP);
   sDisCTSFlowCtl(ChP);
   sDisIXANY(ChP);

   if(ModemCtl & COM_MDM_RTS)          /* turn off modem control lines */
   {
      sClrRTS(ChP);
   }
   if(ModemCtl & COM_MDM_DTR)
   {
      sClrDTR(ChP);
   }

   aaEnableInt();                      /* enable interrupts */
   return(NO_ERR);

} /* end aaClose */

/******************************************************************************
Function: aaSetCloseDelay
Purpose:  Set the maximum time aaClose() will wait for a device's transmit
          buffer to drain before flushing the transmit buffer and
          completing the close.
Call:     aaSetCloseDelay(Dev,MaxDelay)
          int Dev: Device number
          int MaxDelay: Maximum time aaClose will wait for a device's transmit
                        buffer to drain in seconds.  For no delay use 0.
                        Maximum value is 32,767 seconds.
Return:   int: NO_ERR if successful.
               ERR_DEV if device number out of range.
Comments: The device does not need to be open to execute this function.
*/
int aaSetCloseDelay(int Dev,int MaxDelay)
{
   DEV_T *DevP;                    /* ptr to device structure */

   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   DevP = aaDevices[Dev];
   if(MaxDelay < 1)
      DevP->CloseDelay = -1;           /* no delay */
   else
      DevP->CloseDelay = ((long)MaxDelay * 1000)/(TICK_PERIOD/1000);
   return(NO_ERR);
}

/*******************************************************************************
Function: aaGetModemStatus
Purpose:  Get a device's modem status.
Call:     aaGetModemStatus(Dev)
          int Dev: Device number
Return:   unsigned char: State of the modem control inputs using flags
                         COM_MDM_CTS, COM_MDM_DSR, and COM_MDM_CD.  If
                         a flag is set that modem line is ON, if a flag
                         is not set that modem line is OFF.
*/
unsigned char aaGetModemStatus(int Dev)
{
   return(sGetModemStatus(&aaDevices[Dev]->Chan));
} /* end aaGetModemStatus */

/*******************************************************************************
Function: aaWrite
Purpose:  Write serial data out a device.
Call:     aaWrite(Dev,Cnt,Buf)
          int Dev: Device number
          int Cnt: Number of bytes to be written
          unsigned char *Buf: Buffer of data to write
Return:   int: Number of bytes written if successful
               0 if no data bytes written
               ERR_DEV if device number out of range
               ERR_NOTOPEN if device not open for transmit
Warnings: The Cnt parameter should not be greater than the size of the
          transmit buffer Buf.
*/
int aaWrite(int Dev,int Cnt,unsigned char *Buf)
{
   DEV_T *DevP;                    /* ptr to device structure */

   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);

   DevP = aaDevices[Dev];
   if((DevP->Status & (COM_OPEN | COM_TX)) != (COM_OPEN | COM_TX))
   {
      return(ERR_NOTOPEN);
   }
   return(sWriteTxBlk(&DevP->Chan,Buf,Cnt));
} /* end aaWrite */

/*******************************************************************************
Function: aaRead
Purpose:  Read serial data from a device.
Call:     aaRead(Dev,Cnt,Buf)
          int Dev: Device number
          int Cnt: Maximum number of bytes that can be read
          unsigned char *Buf: Buffer to store the data in
Return:   int: Number of bytes read if successful
               0 if no data available to be read
               ERR_DEV if device number out of range
               ERR_NOTOPEN if device not open for receive
Comments: This function reads data from the device's receive buffer
          without checking for receive errors.  If receive error information
          is needed use functions aaGetRxStatus() and aaReadWithStatus().
Warnings: The Cnt parameter should not be greater than the size of the receive
          buffer Buf.
*/
int aaRead(int Dev,int Cnt,unsigned char *Buf)
{
   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */
   int ActualCnt;                      /* actual number of bytes read */
   WORD ChanStatus;                  /* channel status */

   if(Dev < 0 || Dev >= aaDevSize)
      return(ERR_DEV);                 /* invalid device */

   DevP = aaDevices[Dev];
   if((DevP->Status & (COM_OPEN | COM_RX)) != (COM_OPEN | COM_RX))
   {
      return(ERR_NOTOPEN);             /* device not open for read */
   }
   ChP = &DevP->Chan;
   aaDisableInt();                     /* no other AIOP access till done */
   ChanStatus = sGetChanStatus(ChP) &  /* save channel status */
               (STATMODE|RXFOVERFL|RXBREAK|RXFRAME|RXPARITY);
   if(ChanStatus & STATMODE)           /* in Rx status mode */
      sDisRxStatusMode(ChP);           /* disable status mode for this read */
   ActualCnt = sReadRxBlk(ChP,Buf,Cnt); /* read */

#ifndef INTERNAL_ISFLOW
   if(DevP->Status & COM_FLOW_IS)   /* input software flow ctl enabled */
   {
      if(!(DevP->Status & COM_RXFLOW_ON)) /* Rx data flow stopped */
      {
         if(sGetRxCnt(ChP) <= RX_LOWATER) /* time to restart Rx input */
         {
            DevP->Status |= COM_RXFLOW_ON;
            while(!sWriteTxPrioByte(ChP,0x11)); /* send XON */
         }
      }
   }
#endif

   /* If the Rx FIFO is not empty after the read there may still be errors
      somewhere in it.  We must enable Rx status mode again to indicate
      that there are errors in the FIFO.  This is because the earlier
      call to sGetChanStatus() cleared all error bits in the channel status
      register. */
   if((ChanStatus != 0) && (sGetChanStatusLo(ChP) & RDA))
      sEnRxStatusMode(ChP);
   aaEnableInt();
   return(ActualCnt);
} /* end aaRead */

/*******************************************************************************
Function: aaReadWithStatus
Purpose:  Read serial data and status from a device.
Call:     aaReadWithStatus(Dev,Cnt,Buf)
          int Dev: Device number
          int Cnt: Maximum number of bytes that can be read
          WORD *Buf: Buffer to store the data and status.  The low
             byte of each array element in Buf contains the data byte, and the
             high byte contains the status for that data byte.  The status
             may be 0 indicating no error, or any combination of the
             following flags:
                ERR_PARITY:  parity error
                ERR_OVRRUN:  receiver over run error
                ERR_FRAME:   framing error
                ERR_BREAK:   break
Return:   int: Number of bytes read if successful.
               0 if no data available to be read.
               ERR_DEV if device number out of range.
               ERR_NOTOPEN if device not open for receive
Warnings: The Cnt parameter should not be greater than the number of array
          elements in the receive buffer Buf.
*/
int aaReadWithStatus(int Dev,int Cnt,WORD *Buf)
{
   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */
   int RetCount;                       /* actual number of bytes read */
   WORD ChanStatus;                  /* channel status */

   if(Dev < 0 || Dev >= aaDevSize)
      return(ERR_DEV);                 /* invalid device */
   DevP = aaDevices[Dev];
   if((DevP->Status & (COM_OPEN | COM_RX)) != (COM_OPEN | COM_RX))
   {
      return(ERR_NOTOPEN);             /* device not open for read */
   }
   ChP = &DevP->Chan;
   aaDisableInt();             /* no other AIOP access allowed till done */
   RetCount = sGetRxCnt(ChP);          /* number bytes in Rx FIFO */
   if(RetCount <= 0)                   /* no data available */
   {
      aaEnableInt();
      return(0);
   }
   if(RetCount > Cnt)                  /* more available than requested */
      RetCount = Cnt;                  /* only dequeue as much as requested */
   else                                /* request <= available */
      Cnt = RetCount;                  /* only get as much as is available */
   ChanStatus = sGetChanStatus(ChP) &  /* save channel status */
               (STATMODE|RXFOVERFL|RXBREAK|RXFRAME|RXPARITY);
   sEnRxStatusMode(ChP);               /* enable status mode */
   while(Cnt-- > 0)                    /* read */
      *Buf++ = sReadRxWord(sGetTxRxDataIO(ChP));
#ifndef INTERNAL_ISFLOW
   if(DevP->Status & COM_FLOW_IS)      /* input software flow ctl enabled */
   {
      if(!(DevP->Status & COM_RXFLOW_ON)) /* Rx data flow stopped */
      {
         if(sGetRxCnt(ChP) <= RX_LOWATER) /* time to restart Rx input */
         {
            DevP->Status |= COM_RXFLOW_ON;
            while(!sWriteTxPrioByte(ChP,0x11)); /* send XON */
         }
      }
   }
#endif

   /* If the Rx FIFO is not empty after the read there may still be errors
      somewhere in it.  We must leave Rx status mode enabled to indicate
      that there are errors in the FIFO.  This is because the earlier
      call to sGetChanStatus() cleared all error bits in the channel status
      register. */
   if((ChanStatus == 0) || !(sGetChanStatusLo(ChP) & RDA))
   {
      sDisRxStatusMode(ChP);
   }
   aaEnableInt();
   return(RetCount);
} /* end aaReadWithStatus */

/*******************************************************************************
Function: aaGetRxCount
Purpose:  Get the count of data bytes available in the receive buffer.
Call:     aaGetRxCount(Dev)
          int Dev: Device number
Return:   int: Receive byte count
*/
int aaGetRxCount(int Dev)
{
   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   return(sGetRxCnt(&aaDevices[Dev]->Chan));
} /* end aaGetRxCount */

/*******************************************************************************
Function: aaGetRxStatus
Purpose:  Get the status of the device's receive buffer
Call:     aaGetRxStatus(Dev)
          int Dev: Device number
Return:   int: NO_ERR if there are no errors in the device's receive buffer
               ERR_RX if there are errors in the device's receive buffer
               ERR_DEV if device number out of range
               ERR_NOTOPEN if device not open for receive
Comments: If there are errors in the device's receive buffer the exact
          error and the errored data byte can be determined by reading with
          aaReadWithStatus().
*/
int aaGetRxStatus(int Dev)
{
   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */
   WORD ChanStatus;                  /* channel status */

   if(Dev < 0 || Dev >= aaDevSize)
      return(ERR_DEV);                 /* invalid device */
   DevP = aaDevices[Dev];
   if((DevP->Status & (COM_OPEN | COM_RX)) != (COM_OPEN | COM_RX))
   {
      return(ERR_NOTOPEN);             /* device not open for read */
   }
   ChP = &DevP->Chan;
   ChanStatus = sGetChanStatus(ChP) &  /* get channel status */
               (STATMODE|RXFOVERFL|RXBREAK|RXFRAME|RXPARITY|RDA);

   /* If the Rx FIFO is not empty there may still be errors
      somewhere in it.  If so we must leave Rx status mode enabled to
      indicate that there are errors in the FIFO.  This is because the
      earlier call to sGetChanStatus() cleared all error bits in the
      channel status register. */
   if(((ChanStatus & ~RDA) != 0) && (ChanStatus & RDA))
   {
      sEnRxStatusMode(ChP);
      return(ERR_RX);
   }
   return(NO_ERR);
} /* end aaGetRxStatus */

/*---------------------------------------------------------------------------
Function: aaGetRxStatusWord
Purpose:  Get the status WORD of the device's
Call:     aaGetRxStatusWord(Dev)
          int Dev: Device number
Return:   int:  Returns with any or none of the following bit flags set:
            ERR_PARITY, ERR_OVRRUN, ERR_FRAME, ERR_BREAK.

Comments: If there are errors in the device's receive buffer the exact
          error and the errored data byte can be determined by reading with
          aaReadWithStatus().
|---------------------------------------------------------------------------*/
int aaGetRxStatusWord(int Dev)
{
 DEV_T *DevP;                    /* ptr to device structure */
 CHANPTR_T ChP;                      /* ptr to channel structure */
 WORD ChanStatus;                  /* channel status */
 int ret_stat;

   if(Dev < 0 || Dev >= aaDevSize)
      return 0;                 /* invalid device */
   DevP = aaDevices[Dev];
   if((DevP->Status & (COM_OPEN | COM_RX)) != (COM_OPEN | COM_RX))
   {
      return 0;             /* device not open for read */
   }
   ChP = &DevP->Chan;
   ChanStatus = sGetChanStatus(ChP);  /* get channel status */

   if (0 != ChanStatus)  // if error or break
   {
     ret_stat = 0;
     if (ChanStatus & RXPARITY)
       ret_stat |= ERR_PARITY;
     if (ChanStatus & RXFOVERFL)
       ret_stat |= ERR_OVRRUN;
     if (ChanStatus & RXFRAME)
       ret_stat |= ERR_FRAME;
      if (ChanStatus & RXBREAK)
       ret_stat |= ERR_BREAK;

     return ret_stat;  // error or break
   }
   return 0;  // nothing unusual

} /* end aaGetRxStatusWord */

/*******************************************************************************
Function: aaGetTxCount
Purpose:  Get the count of data bytes in the transmit buffer waiting to
          be transmitted.
Call:     aaGetTxCount(Dev)
          int Dev: Device number
Return:   int: Transmit byte count
*/
int aaGetTxCount(int Dev)
{
   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   return((int)sGetTxCnt(&aaDevices[Dev]->Chan));
} /* end aaGetTxCount */

/*******************************************************************************
Function: aaChangeModemState
Purpose:  Change the state of modem output lines.
Call:     aaChangeModemState(Dev,RTSState,DTRState)
          int Dev;       Device number
          int RTSState:  State of RTS line: ON, OFF, or NOCHANGE
          int DTRState:  State of DTR line: ON, OFF, or NOCHANGE
Return:   int: NO_ERR if successful
               ERR_DEV if device out of range
Warnings: Setting RTS if the device was opened for input hardware flow
          control (flag COM_FLOW_IH) will override RTS operation for
          flow control.
*/
int aaChangeModemState(int Dev,int RTSState,int DTRState)
{
   CHANPTR_T ChP;

   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   ChP = &aaDevices[Dev]->Chan;
   if(RTSState == OFF)
   {
      sClrRTS(ChP);
   }
   else if(RTSState == ON)
   {
      sSetRTS(ChP);
   }
   if(DTRState == OFF)
   {
      sClrDTR(ChP);
   }
   else if(DTRState == ON)
   {
      sSetDTR(ChP);
   }
   return(NO_ERR);
} /* end aaChangeModemState */

/******************************************************************************
Function: aaFlush
Purpose:  Flush transmit and/or receive buffer for a device.
Call:     aaFlush(Dev,FlushFlags)
          int Dev: Device number
          unsigned char FlushFlags: COM_TX and/or COM_RX
Return:   int: NO_ERR if successful
               ERR_DEV if device is out of range
               ERR_OPENTYPE if FlushFlags is out of range
*/
int aaFlush(int Dev,unsigned char FlushFlags)
{
   CHANPTR_T ChP;
   DEV_T *DevP;                    /* ptr to device structure */
   unsigned char PrioData;             /* next byte in Tx prio buf */

   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   if(aaCheckOpenType(FlushFlags) < 0) /* check flush flags */
      return(ERR_OPENTYPE);
   DevP = aaDevices[Dev];
   ChP = &DevP->Chan;

   aaDisableInt();                 /* no other AIOP access allowed till done */
   if(FlushFlags & COM_TX)             /* flush Tx */
   {
      sFlushTxFIFO(ChP);
   }
   if(FlushFlags & COM_RX)             /* flush Rx */
   {
      sFlushRxFIFO(ChP);
#ifndef INTERNAL_ISFLOW
      if(DevP->Status & COM_FLOW_IS)   /* input software flow ctl enabled */
      {
         if(!(DevP->Status & COM_RXFLOW_ON)) /* Rx data flow stopped */
         {
            DevP->Status |= COM_RXFLOW_ON;
            while(!sWriteTxPrioByte(ChP,0x11)); /* send XON */
         }
      }
#endif
   }
   aaEnableInt();
   return(NO_ERR);
}

/******************************************************************************
Function:   aaExit
Purpose:    Perform cleanup when exiting from an application program.
Call:       aaExit()
Return:     void
Comments:   This function will do any cleanup tasks required when exiting
            from an application, such as halting controller periodic
            interrupts, restoring the IRQ vector used by the controller,
            and resetting all serial devices.
Warnings:   Once aaInit() has been called, aaExit() must be called before
            exiting the application program.

            If the application program can be exited using the CTRL+C or
            CTRL+BREAK keys, then the default DOS CTRL+C handler must be
            replaced with a handler that calls aaExit().  Function
            aaInstallCtrlCHandler() can be used for this purpose.
*/
void aaExit(void)
{
   static int Ctl;                     /* controller number */
   static int Dev;                     /* device number */
   static int LastDev;                 /* last device number on controller */

   sDisGlobalInt(&sController[0]);
   aaRestoreIrq(aaIrqSave);
   for(Ctl = 0;Ctl < aaCtlSize;Ctl++)  /* reset all serial devices */
   {
      Dev = aaCfg.CfCtl[Ctl].StartDev;
      if(Dev == -1)                    /* this controller not installed */
         break;
      LastDev = Dev + aaCfg.CfCtl[Ctl].NumChan;
      for(;Dev < LastDev;Dev += 8)
         sResetAiop(&aaDevices[Dev]->Chan);
   }
}

/******************************************************************************
Function: aaSendBreak
Purpose:  Sends a break signal.
Call:     aaSendBreak(Dev,Time)
          int Dev: Device number
          int Time: Time in milliseconds to send the break
Return:   int: NO_ERR if successful
               ERR_DEV if device is out of range.
*/
int aaSendBreak(int Dev,int Time)
{
   DEV_T *DevP;

   if(Dev < 0 || Dev >= aaDevSize)     /* check device bounds */
      return(ERR_DEV);
   DevP = aaDevices[Dev];

   DevP->Status |= COM_BREAK_ON;
   DevP->BreakTime = (long)Time * 1000; /* break time in microseconds */
   sSendBreak(&DevP->Chan);
   return(NO_ERR);
} /* end aaSendBreak */

/*******************************************************************************
Function: aaCheckOpenType
Purpose:  Check for a valid Open Type.
Call:     aaCheckOpenType(OpenType)
          WORD OpenType: Opening for read or write
Return:   int: NO_ERR if open type valid, ERR_OPENTYPE if out of range.
Comments: This function checks the 'open type' parameter for an aaOpen()
          function call.
Warning:  This is an API privelaged function
*/
int aaCheckOpenType(WORD OpenType)
{
  int i;

  for (i=0; i<MAXNUMOTS; i++) /* Check Open Type */
  {
     if (aaOpenTypeCheck[i] == OpenType)
        return(NO_ERR);
  }
  return(ERR_OPENTYPE);
}

/*******************************************************************************
Function: aaCheckBaudRate
Purpose:  Check for a valid Baud Rate.
Call:     aaCheckBaudRate(Baud)
          unsigned char Baud: This is the selected baud rate.
Return:   int: NO_ERR if open type valid, ERR_BAUDRATE if out of range.
Comments: This function checks the 'baud rate' parameter for an aaOpen()
          or aaReconfigure() function call.
Warning:  This is an API privelaged function
*/
int aaCheckBaudRate(unsigned long Baud)
{
 long newbaud;

  if (Baud < MAXNUMBAUDS)
        newbaud = aaBaud[(int)Baud];
  else  newbaud = Baud;

  /* we simply check the baudrate here(3rd param=0) */
  if (sSetBaudRate(NULL, newbaud, 0))  /* check the baud */
    return(ERR_BAUDRATE);

  return(NO_ERR);
}

/*******************************************************************************
Function: aaCheckParity
Purpose:  Check for valid Parity.
Call:     aaCheckParity(Parity)
          BYTE_T Parity: This is the selected parity
Return:   int: NO_ERR if open type valid, ERR_PAR if out of range.
Comments: This function checks the 'parity' parameter for an aaOpen()
          or aaReconfigure() function call.
Warning:  This is an API privelaged function
*/
int aaCheckParity(unsigned char Parity)
{
  int i;

  for (i=0; i<MAXNUMPARITYS; i++) /* Check Parity */
  {
     if (aaParityCheck[i] == Parity)
        return(NO_ERR);
  }
  return(ERR_PAR);
}

/*******************************************************************************
Function: aaCheckDataB
Purpose:  Check for valid Data Bits.
Call:     aaCheckDataB(Datab)
          unsigned char Datab: This is the selected data bits
Return:   int: NO_ERR if open type valid, ERR_DATAB if out of range.
Comments: This function checks the 'data bits' parameter for an aaOpen()
              or aaReconfigure() function call.
Warning:  This is an API privelaged function
*/
int aaCheckDataB(unsigned char Datab)
{
  int i;

  for (i=0; i<MAXNUMDATABS; i++) /* Check Data bits */
  {
     if (aaDataBCheck[i] == Datab)
        return(NO_ERR);
  }
  return(ERR_DATAB);
}

/*******************************************************************************
Function: aaCheckStopB
Purpose:  Check for valid Stop Bits.
Call:     aaCheckStopB(Stopb)
          unsigned char Stopb: This is the selected data bits
Return:   int: NO_ERR if open type valid, ERR_STOPB if out of range.
Comments: This function checks the 'stop bits' parameter for an aaOpen()
          or aaReconfigure() function call.
Warning:  This is an API privelaged function
*/
int aaCheckStopB(unsigned char Stopb)
{
  int i;

  for (i=0; i<MAXNUMSTOPBS; i++) /* Check Stop bits */
  {
     if (aaStopBCheck[i] == Stopb)
        return(NO_ERR);
  }
  return(ERR_STOPB);
}


/*******************************************************************************
Function: aaInstallIrq
Purpose:  Install an Interrupt Request vector and enable it.
Call:     aaInstallIrq(Irq,IntHnd)
          int Irq; Interrupt request line number
          void (interrupt far *IntHnd)(); Ptr to interrupt handler
Return:   void
Comments: This functions saves the old interrupt vector, installs the new
          vector, saves the previous state of the 8259 mask register bit for
          this IRQ, and enables this IRQ in the 8259 mask register.
Warnigs:  This function disables and reenables interrupts.
          This is an API privelaged function
*/
void aaInstallIrq(int Irq,void (interrupt far *IntHnd)())
{
   IRQ_T far *IrqP;                    /* ptr to IRQ information structure */
   unsigned char Mask;                 /* 8259 mask */

   aaIrqSave = Irq;                    /* save this IRQ number */

   if (Irq == 0)  // run with no irq, 4-02-96 add
     return;

   IrqP = &aaIrqInfo[Irq];             /* get ptr to IRQ info structure */
   aaEndOfIntIO = IrqP->MaskIO - 1;    /* save EOI address */

   /* Save vector.  The cast is needed to avaoid a C 7.0 warning that
      I can't get rid of otherwise. */
   aaVectorSave = (void (interrupt far *)(void))_dos_getvect(IrqP->IntNum);

   aaDisableInt();                     /* disable interrupts */
   _dos_setvect(IrqP->IntNum,IntHnd);  /* put new vector in vector table */
   Mask = sInB(IrqP->MaskIO);          /* get current 8259 mask register */
   aaMaskSave = Mask & ~IrqP->Mask;    /* save bit for IRQ line, others low */
   Mask &= IrqP->Mask;                 /* enable IRQ line */
   sOutB(IrqP->MaskIO,Mask);           /* write it to 8259 */
   aaEnableInt();                      /* enable interrupts */
}

/*******************************************************************************
Function: aaRestoreIrq
Purpose:  Restore an IRQ to it state before it was installed.
Call:     aaRestoreIrq(Irq)
          int Irq; Interrupt request line number
Return:   void
Comments: This functions restores the old interrupt vector and restores
          the previous 8259 mask register bit for this IRQ line.
Warnigs:  This function disables and reenables interrupts.
          This is an API privelaged function
*/
void aaRestoreIrq(int Irq)
{
   IRQ_T far *IrqP;                    /* ptr to IRQ information structure */
   unsigned char Mask;                 /* 8259 mask */

   if(aaIrqSave == 0)                  /* if vector never installed */
      return;                          /* don't reinstall it */
   IrqP = &aaIrqInfo[Irq];             /* get ptr to IRQ info structure */
   aaDisableInt();                     /* disable interrupts */
   _dos_setvect(IrqP->IntNum,aaVectorSave);

   Mask = sInB(IrqP->MaskIO);          /* get current 8259 mask register */
   Mask |= aaMaskSave;                 /* restore bit for IRQ, others same */
   sOutB(IrqP->MaskIO,Mask);           /* write it to 8259 */
   aaIrqSave = 0;                      /* indicate our vector not installed */
   aaEnableInt();                      /* enable interrupts */
}

/*******************************************************************************
Function: aaIntHandler
Purpose:  Interrupt service routine
Call:     aaIntHandler()
Return:   void
Comments: This interrupt handler reenables interrupts immediately
          upon entry so that higher priority interrupt service routines
          can interrupt.  However, the EOI to the 8259 is not done till
          the end of the handler, so no additional RocketPort interrupts
          should occur till all processing is done.
Warning:  This is an API privelaged function
*/
void interrupt far aaIntHandler(void)
{
   int Ctl;                            /* controller number */
   int Dev;                            /* device number */
   int Aiop;                           /* AIOP number */
   int Chan;                           /* channel number within AIOP */
   CONTROLLER_T *CtlP;                 /* ptr to controller structure */
   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */
   register WORD Status;             /* device status */
   unsigned char CtlInt;               /* controller interrupt status */
   unsigned char AiopInt;              /* AIOP interrupt status */
   unsigned char IntID;                /* channel interrupt ID */

   aaControllerTime += TICK_PERIOD;    /* bump controller time */
   aaEnableInt();                      /* enable interrupts */

   for(Dev = 0;Dev < aaDevSize;Dev++)  /* check each device for work */
   {
      /* Set up device and channel structure pointers */
      DevP = aaDevices[Dev];
      Status = DevP->Status;
      if(!(Status & COM_OPEN))
         continue;
      ChP = &DevP->Chan;

      /* Break processing */
      if(Status & COM_BREAK_ON)
      {
         if(DevP->BreakTime <= 0)
         {
            sClrBreak(ChP);
            DevP->Status &= ~COM_BREAK_ON;
         }
         else
         {
            DevP->BreakTime -= TICK_PERIOD;
         }
      }

      /* Input software flow control */
#ifndef INTERNAL_ISFLOW
      if(Status & COM_FLOW_IS)         /* input software flow ctl enabled */
      {
         if(Status & COM_RXFLOW_ON)    /* Rx data coming in */
         {
            if(sGetRxCnt(ChP) >= RX_HIWATER) /* time to stop Rx input */
            {
               DevP->Status &= ~COM_RXFLOW_ON;
               while(!sWriteTxPrioByte(ChP,0x13)); /* send XOFF */
            }
         }
      }
#endif
   }

   /* Dispatch event functions */
   if(aaPeriodicEventOn)
   {
      (*aaPeriodicEventP)();
   }

   for(Ctl = 0;Ctl < aaCtlSize;Ctl++)
   {
      CtlP = &sController[Ctl];
      CtlInt = sGetControllerIntStatus(CtlP);
      while((Aiop = sGetNextInBitMap(CtlInt)) != -1)
      {
         sClrNextInBitMap(CtlInt,Aiop);
         AiopInt = sGetAiopIntStatus(CtlP,Aiop);
         while((Chan = sGetNextInBitMap(AiopInt)) != -1)
         {
            sClrNextInBitMap(AiopInt,Chan);
            Dev = sGetDevMap(Ctl,Aiop,Chan);
            ChP = &aaDevices[Dev]->Chan;
            IntID = sGetChanIntID(ChP);
            if(IntID & RXF_TRIG)
            {
               (*aaRxEventP)(Dev);
            }
            IntID &= DELTA_CD|DELTA_CTS|DELTA_DSR;
            if(IntID)
            {
               (*aaMdmChgEventP)(Dev,aaModemChg[IntID],sGetModemStatus(ChP));
            }
         }
      }
   }

   aaDisableInt(); /* disable ints so can completely exit ISR without interruption */
   sOutB(aaEndOfIntIO,0x20);           /* write nonspecific EOI */
   if(aaEndOfIntIO == 0xa0)            /* was not master 8259 */
   {
      sOutB(0x20,0x20);                /* write nonspecific EOI to master */
   }

   if ((&sController[0])->BusType != BUS_PCI)  // mudback does not exist for PCI */
     sControllerEOI(&sController[0]); /* EOI to controller generating int */
   else
     sPCIControllerEOI(CtlP);               /* clear EOI if warm init */
}

/*******************************************************************************
Function: aaPollHandler
Purpose:  service routine
Call:     aaPollHandler()
Return:   void
Comments: This is called by the interrupt routine if used.  Otherwise the
          program can poll it if not using interrupts.
*/
void aaPollHandler(int tick_period)
{
   int Ctl;                            /* controller number */
   int Dev;                            /* device number */
   int Aiop;                           /* AIOP number */
   int Chan;                           /* channel number within AIOP */
   CONTROLLER_T *CtlP;                 /* ptr to controller structure */

   DEV_T *DevP;                    /* ptr to device structure */
   CHANPTR_T ChP;                      /* ptr to channel structure */
   register WORD Status;             /* device status */
   unsigned char CtlInt;               /* controller interrupt status */
   unsigned char AiopInt;              /* AIOP interrupt status */
   unsigned char IntID;                /* channel interrupt ID */

   aaControllerTime += tick_period;    /* bump controller time */

   for(Dev = 0;Dev < aaDevSize;Dev++)  /* check each device for work */
   {
      /* Set up device and channel structure pointers */
      DevP = aaDevices[Dev];
      Status = DevP->Status;
      if(!(Status & COM_OPEN))
         continue;
      ChP = &DevP->Chan;

      /* Break processing */
      if(Status & COM_BREAK_ON)
      {
         if(DevP->BreakTime <= 0)
         {
            sClrBreak(ChP);
            DevP->Status &= ~COM_BREAK_ON;
         }
         else
         {
            DevP->BreakTime -= TICK_PERIOD;
         }
      }

      /* Input software flow control */
#ifndef INTERNAL_ISFLOW
      if(Status & COM_FLOW_IS)         /* input software flow ctl enabled */
      {
         if(Status & COM_RXFLOW_ON)    /* Rx data coming in */
         {
            if(sGetRxCnt(ChP) >= RX_HIWATER) /* time to stop Rx input */
            {
               DevP->Status &= ~COM_RXFLOW_ON;
               while(!sWriteTxPrioByte(ChP,0x13)); /* send XOFF */
            }
         }
      }
#endif
   }

   /* Dispatch event functions */
   if(aaPeriodicEventOn)
   {
      (*aaPeriodicEventP)();
   }

   for(Ctl = 0;Ctl < aaCtlSize;Ctl++)
   {
      CtlP = &sController[Ctl];
      CtlInt = sGetControllerIntStatus(CtlP);
      while((Aiop = sGetNextInBitMap(CtlInt)) != -1)
      {
         sClrNextInBitMap(CtlInt,Aiop);
         AiopInt = sGetAiopIntStatus(CtlP,Aiop);
         while((Chan = sGetNextInBitMap(AiopInt)) != -1)
         {
            sClrNextInBitMap(AiopInt,Chan);
            Dev = sGetDevMap(Ctl,Aiop,Chan);
            ChP = &aaDevices[Dev]->Chan;
            IntID = sGetChanIntID(ChP);
            if(IntID & RXF_TRIG)
            {
               (*aaRxEventP)(Dev);
            }
            IntID &= DELTA_CD|DELTA_CTS|DELTA_DSR;
            if(IntID)
            {
               (*aaMdmChgEventP)(Dev,aaModemChg[IntID],sGetModemStatus(ChP));
            }
         }
      }
   }
}

/*******************************************************************************
Function: aaCtrlCIntHandler
Purpose:  CTRL+C Interrupt service routine
Call:     aaCtrlCIntHandler()
Return:   void
Comments: If the operator hits the CTRL+C the default DOS CTRL+C handler
          will terminate the program without allowing it to go through
          aaExit().  This results in a system hang because RocketPort
          periodic interrupts continue to run after the interrupt code
          has been removed.

          This handler corrects this problem by calling aaExit()
          to restore the IRQ vector and stop Rocket periodic interrupts.

          It's OK for this handler to call DOS system functions because
          DOS cleans up before issuing the INT 23H - see MS-DOS Prog. Ref.
Warning:  This is an API privelaged function.
          If the user wants to install his own CTRL+C handler after aaInit()
          is called he must chain it to this handler.
*/
void far aaCtrlCIntHandler(void)
{
   aaExit();
   aaSetCarry();
}


/******************************************************************************
Function: aaInstallMdmChgEvent
Purpose:  Installs an application level event function handle modem change
          events.
Call:     aaInstallMdmChgEvent(EvFuncP)
          void (*EvFuncP)(Dev,unsigned char MdmChange,unsigned char MdmState);
             Ptr to the event function
Return:   void
Comments: See function EvModemChange() for a description of the event function.
Warnings: The function installed here will be called during an ISR.  Keep
          your code short and remember that many standard 'C' library calls
          do not work in ISRs, such as printf().

          If using the Microsoft C compiler stack checking must be disabled
          during the event function and any functions called by the event
          function.  Stack checking can be turned off with
             #pragma check_stack(off)
          and turned back on with
             #pragma check_stack(on)

          If interrupts are not used, the aaPollHandler() routine should
          be called by your application at regular intervals to support
          this function.
*/
void aaInstallMdmChgEvent(void (*EvFuncP)(int,unsigned char,unsigned char))
{
   aaMdmChgEventP = (EvFuncP);     /* point to application's event function */
}

/******************************************************************************
Function: aaInstallRxEvent
Purpose:  Installs an application level event function to handle receive data
          available events.
Call:     aaInstallRxEvent(EvFuncP)
          void (*EvFuncP)(int Dev); Ptr to the event function.
Return:   void
Comments: See function EvRxData() for a description of the event function.
Warnings: The function installed here will be called during an ISR.  Keep
          your code short and remember that many standard 'C' library calls
          do not work in ISRs, such as printf().

          If using the Microsoft C compiler stack checking must be disabled
          during the event function and any functions called by the event
          function.  Stack checking can be turned off with
             #pragma check_stack(off)
          and turned back on with
             #pragma check_stack(on)
          If interrupts are not used, the aaPollHandler() routine should
          be called by your application at regular intervals to support
          this function.
*/
void aaInstallRxEvent(void (*EvFuncP)(int))
{
   aaRxEventP=(EvFuncP);           /* point to application's event function */
}

/******************************************************************************
Function: aaInstallPeriodicEvent
Purpose:  Installs a periodic application level event function.
Call:     aaInstallPeriodicEvent(EvFuncP)
          void (*EvFuncP)(void); Ptr to the event function.
Return:   void
Comments: See function EvPeriodic() for a description of the event function.
          The periodic event function is called 274 times a second.  Once
          installed the periodic event function will not be dispatched until
          it is enabled with aaEnPeriodicEvent().  Function aaEnPeriodicEvent()
          can also be used to disable dispatching of the periodic event
          function.
Warnings: The function installed here will be called during an ISR.  Keep
          your code short and remember that many standard 'C' library calls
          do not work in ISRs, such as printf().

          If using the Microsoft C compiler stack checking must be disabled
          during the event function and any functions called by the event
          function.  Stack checking can be turned off with
             #pragma check_stack(off)
          and turned back on with
             #pragma check_stack(on)
          If interrupts are not used, the aaPollHandler() routine should
          be called by your application at regular intervals to support
          this function.
*/
void aaInstallPeriodicEvent(void (*EvFuncP)(void))
{
   aaPeriodicEventP=(EvFuncP);     /* point to application's event function */
}

/******************************************************************************
Function: aaEnPeriodicEvent
Purpose:  Enables or disables dispatching of the periodic event function.
Call:     aaEnPeriodicEvent(State)
          int State; TRUE to enable dispatching of the periodic event function,
                     FALSE to disable dispatching.
Return:   void
Comments: The periodic event function is called 274 times a second.  Once
          installed the periodic event function will not be dispatched until
          it is enabled with aaEnPeriodicEvent().  Function aaEnPeriodicEvent()
          can also be used to disable dispatching of the periodic event
          function.
Warnings: The event function must be installed with aaInstallPeriodicEvent()
          before enabling dispatching.
*/
void aaEnPeriodicEvent(int State)
{
   if(State)
      aaPeriodicEventOn = TRUE;
   else
      aaPeriodicEventOn = FALSE;
}

/******************************************************************************
Function: aaMdmChgEventNull
Purpose:  Define a default modem change event function.
Call:     aaMdmChgEventNull(Dev,MdmChange,MdmState)
              int Dev: Device number
              unsigned char MdmChange: Modem input lines which changed
              unsigned char MdmState:  Current state of modem lines
Return:   void
Comments: This function is called when a modem status change has occurred
          if no application level routine has been installed.
Warnings: This function is called during an ISR.
          This is an API privelaged function.
*/
void aaMdmChgEventNull(int Dev,unsigned char MdmChange,unsigned char MdmState)
{
   Dev = Dev;                          /* prevent compiler warning */
   MdmChange = MdmChange;              /* prevent compiler warning */
   MdmState = MdmState;                /* prevent compiler warning */
}

/******************************************************************************
Function: aaRxEventNull
Purpose:  Define a default receive data event function.
          aaRxEventNull(int Dev)
          int Dev: Device number
Return:   void
Comments: This function is called if receive data is available for a device
          and if no application level routine has been installed.
Warnings: This function called during an ISR.
          This is an API privelaged function.
*/
void aaRxEventNull(int Dev)
{
   Dev = Dev;                          /* prevent compiler warning */
}

/******************************************************************************
Function: aaPeriodicEventNull
Purpose:  Define a default periodic event function.
          aaPeriodicEventNull(int Dev)
          int Dev: Device number
Return:   void
Comments: This function is called if periodic events are enabled but no
          application level event function has been installed.
Warnings: This function called during an ISR.
          This is an API privelaged function.
*/
void aaPeriodicEventNull(void)
{
   ;
}

/******************************************************************************
Function: aaDelay
Purpose:  Delay for a specific amount of time in milliseconds
Call:     void aaDelay(Wait)
          int Wait; time to wait in milliseconds
Return:   void
Comments: This function keeps control of the CPU until the Wait interval
          has elapsed.  Resolution is +8.3 milliseconds.  If Wait == 0
          this function returns immediately.
Warnings: Interrupts must not be disabled when calling this function.
          This is an API privelaged function
*/
void aaDelay(int Wait)
{
   static unsigned long Ticks;                /* number of interrupt ticks to wait */
   static unsigned long StartTime;            /* controller time */
   static long WatchDog;                      /* make sure we don't hang in here */
   static unsigned char x;

   if(Wait <= 0)
      return;

   WatchDog = 0;
   Ticks  = (unsigned long)Wait * 1000 / TICK_PERIOD; /* get num int ticks */
   Ticks++;                            /* for fractional truncation */
   StartTime = aaControllerTime;
   while(StartTime == aaControllerTime) /* wait for time to roll over once */
   {
      x = sGetAiopIntStatus(&sController[0],0); /* use this for a delay */
      x = x;                            /* eliminate compiler warning */
      if(WatchDog++ > 0x300000L)        /* 5 secs on ISA, 3 secs on MC */
      {
         printf("ROCKETPORT INTERRUPT FAILURE [%lx,%lx]\n",StartTime,aaControllerTime);
         return;
      }
   }

   StartTime = aaControllerTime;
   while(Ticks > 0)                    /* count down number of ticks */
   {
      WatchDog = 0;
      while(StartTime == aaControllerTime) /* wait for time to roll over */
      {
         x = sGetAiopIntStatus(&sController[0],0); /* use this for a delay */
         x = x;                        /* eliminate compiler warning */
         if(WatchDog++ > 0x300000L)    /* 5 secs on ISA, 3 secs on MC */
         {
            printf("ROCKETPORT INTERRUPT FAILURE [%lx,%lx]\n",StartTime,aaControllerTime);
            return;
         }
      }
      StartTime = aaControllerTime;
      Ticks--;
   }
}

