/*------------------------------------------------------------------
 rcktldef.h - RocketPort hardware definitions, dependent compiler and
  OS code goes here!

  The sInStrW() and sOutStrW() macros are the main transfer agents to
  and from the RocketPort.  These are good candidates for speed optimization.

  The sOutDW() (output double word) is used to output configuration
  information to the RocketPort.  It can be useful to use the advanced
  286 out instruction setup for one single double-word operation.  We
  don't really need the speed, but the single un-interruptable op-code
  has some advantages to avoid hardware contention issues possible with
  2 word writes.

 Copyright (c) 1995,1996 Comtrol Corporation
|---------------------------------------------------------------------*/
// Choose one of these or define a new one
//#define WATCOMC
//#define BORLANDC
//#define MSC

// Choose one of these(SIMPLE_IO good for porting)
//#define FAST_IO
#define SIMPLE_IO

#define WANT_BASE_TYPES

#ifdef WANT_BASE_TYPES
typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef int BOOL;
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

//#define SET_LOCK   disable()
//#define CLEAR_LOCK enable()
#define SET_LOCK   
#define CLEAR_LOCK 

/*------ Setup for use with WATCOM C/386 and the WATCOM NLM SDK ------*/
#ifdef WATCOMC

#include <conio.h>

#define ASM _asm

#define sOutB(a,b) outp(a,b)
#define sOutW(a,b) outpw(a,b)
#define sOutDW(a,b) outpd(a,b)
#define sInB(a) (inp(a))
#define sInW(a) (inpw(a))

void aaEnableInt(void);
#pragma aux aaEnableInt = "sti"

void aaDisableInt(void);
#pragma aux aaDisableInt = "cli"

void aaSetCarry(void);
#pragma aux aaSetCarry = "stc"

#ifdef FAST_IO
// void sOutStrW(word, WORD *, WORD);
#pragma aux sOutStrW =   \
  "  cld"                \
  "  rep outsw",         \
  parm [ EDX ] [ ESI ] [ ECX ]  \
  modify [ ESI ECX ];

// void sInStrW(word, WORD *, WORD);
#pragma aux sInStrW =  \
  "  cld"              \
  "  rep insw",        \
  parm [ EDX ] [ EDI ] [ ECX ]  \
  modify [ EDI ECX ];
#endif

#endif
/*------ Setup for use with BORLAND C++ 3.1 ---------------------*/
#ifdef BORLANDC

#include <dos.h>

#define ASM asm

#define sOutB(a, b) outp(a, b)
#define sOutW(a, b) outpw(a, b)
#define sInB(a) (inp(a))
#define sInW(a) (inpw(a))

#define aaEnableInt() ASM sti
#define aaDisableInt() ASM cli
#define aaSetCarry() ASM stc

#ifdef FAST_IO
// generate a 386/486 outdw function
#define sOutDW(a, b) \
{\
  _DX= (a);\
  __emit__(0x66);\
  _AX= (b);\
  __emit__(0x66,0xef);\
}

#define sOutStrW(a, b, c) \
{\
  __emit__(0xfc);\
  _DS= FP_SEG((b));\
  _SI= FP_OFF((b));\
  _CX= (c);\
  _DX= (a);\
  __emit__(0xf3,0x6f);\
}

#define sInStrW(a, b, c)\
{\
  __emit__(0xfc);\
  _ES= FP_SEG((b));\
  _DI= FP_OFF((b));\
  _CX= (c);\
  _DX= (a);\
  __emit__(0xf3,0x6d);\
}
#endif

#endif

/*------ Setup for use with Microsoft ---------------------*/
#ifdef MSC

#define ASM _asm
#define sOutB(IO,VAL)     _outp(IO,VAL)
#define sOutW(IO,VAL)     _outpw(IO,VAL)
// #define sOutDW(IO,VAL) _outpd(IO,VAL)
#define sInB(IO)  (_inp(IO))
#define sInW(IO) (_inpw(IO))

#define aaEnableInt() ASM sti
#define aaDisableInt() ASM cli
#define aaSetCarry() ASM stc
#endif
/*------ End Compiler specific Macros ---------------------*/


//------ Common simple IO macros, good for porting where fast io macros
//------ can be compiler dependent and model dependent.
#ifdef SIMPLE_IO

#ifndef WATCOMC
#define sOutDW(a, b) \
{\
  sOutW(a, (unsigned int) b);\
  sOutW(a+2,(unsigned int) (b >> 16));\
}
#endif

#define sOutStrW(IO,BUFFER,COUNT) \
{ \
   while((COUNT)-- > 0) \
      sOutW(IO,*(BUFFER)++); \
}

// this definition needs a WORD ptr!!!!!!
#define sInStrW(IO,BUFFER,COUNT) \
{ \
   while((COUNT)-- > 0) \
      *(BUFFER)++ = sInW(IO); \
}

#endif
