/*-----------------------------------------------------------------------------
 install.c - Sample install routine to setup config files.
|-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <dos.h>
#include <process.h>
#include <stdlib.h>
#include <conio.h>
#include <ctype.h>
#include "api.h"
// #include "pcibios.h"

int pci_setup(void);
int isa_setup(void);
int yesno(char *str);
int get_num(char *str, unsigned int *wi, int hex);

DEVINFO brds[4];
int total_boards = 0;
int irq_to_use = 0;

/*-----------------------------------------------------------------------------
  main
|-----------------------------------------------------------------------------*/
main(int argc, char **argv)
{
  // char tmpstr[160];
  int stat,i;
  FILE *fp;
  int end_pci = 0;
  unsigned int wnum;

  {
    total_boards = 0;
    irq_to_use = 0;

    printf("Complete the following questions to configure your RocketPort Boards\n");

    stat = yesno("Do you have RocketPort PCI boards(y/n)?:");
    if (stat == 'Y')
    {
      pci_setup();
      printf("%d RocketPort PCI boards found.\n", total_boards);
      end_pci = total_boards;  // keep track of how many
    }

    //printf("\nInstalling RocketPort board %d\n", total_boards+1);

    if (total_boards < 4)
    {
      stat = yesno("Do you have RocketPort ISA boards(y/n)?:");
      if (stat == 'Y')
      {
        isa_setup();
        printf("%d RocketPort ISA boards configured.\n", total_boards - end_pci);
      }
    }

    if (total_boards > 0)
    {
      stat = yesno("Does your program use RocketPort hardware interrupts?");
      if (stat == 'Y')
      {
        if (end_pci != 0)  // pci-boards found
        {
          if (brds[0].Irq == 0)
          {
            // complain about lack of PCI interrupt assigned.
            printf("The First PCI board is not assigned an interrupt.\n");
            printf("Check your computer BIOS setup.\n");
          }
        }
        else
        {
          stat = get_num("Interrupt to use(0=none,11,10,9,7,5,4,3,12):", &wnum,0);
          if (stat == 0)
          {
            brds[0].Irq = (int) wnum;
            printf("IRQ %d selected\n",brds[0].Irq);
          }
        }
      }
      else brds[0].Irq = 0;
    }

    printf("You have configured %d RocketPort boards:\n",total_boards);
    if (brds[0].Irq != 0)
    {
      printf("IRQ %d.\n",brds[0].Irq);
    }
    for (i=0; i< total_boards; i++)
    {
      if (brds[i].Base_Addr != 0)
      {
        printf("%d) IO Address:%xH\n",i+1, brds[i].Base_Addr);
      }
    }

    fp = fopen("config.dat", "w");
    if (brds[0].Base_Addr != 0)
    {
      fprintf(fp, "%d\n", brds[0].Irq);
      fprintf(fp, "%x %d\n", brds[0].Base_Addr, brds[0].num_ports);
      if (brds[1].Base_Addr != 0)
      {
        fprintf(fp, "%x %d\n", brds[1].Base_Addr, brds[1].num_ports);
        if (brds[2].Base_Addr != 0)
        {
          fprintf(fp, "%x %d\n", brds[2].Base_Addr, brds[2].num_ports);
          if (brds[3].Base_Addr != 0)
          {
            fprintf(fp, "%x %d\n", brds[3].Base_Addr, brds[3].num_ports);
          }
        }
      }
    }

    printf("The CONFIG.DAT file has been saved.\n");
    printf("Copy this file to your root directory\n");
    printf("and then add:\n");
    printf("  SET ROCKETCFG=c:\\CONFIG.DAT\n");
    printf("to your AUTOEXEC.BAT file.\n");
  }
  exit(0);
  return 0;
}

#ifdef COMMENT_OUT
/*-----------------------------------------------------------------------------
 show_config -
|-----------------------------------------------------------------------------*/
int show_config(void)
{
 char *str;
 FILE *fp;
 int i;
 unsigned int brd_addr[4];
 int irq_to_use;

  str = getenv("ROCKETCFG");
  if (str == NULL)
    return 1;

  fp = fopen(str,"r");
  if (fp == NULL)
    return 2;

  i = fscanf(fp,"%d%x%x%x%x",
              &irq_to_use,
              &brd_addr[0],
              &brd_addr[1],
              &brd_addr[2],
              &brd_addr[3]);
}
#endif

/*-----------------------------------------------------------------------------
 get_num -
|-----------------------------------------------------------------------------*/
int get_num(char *str, unsigned int *wi, int hex)
{
 int nums;

  printf(str);
  if (hex)
    nums = scanf("%x", wi);
  else
    nums = scanf("%d", wi);
  printf("\n");
  if (nums == 1)
       return 0;  // ok
  else return 1;
}

/*-----------------------------------------------------------------------------
 yesno -
|-----------------------------------------------------------------------------*/
int yesno(char *str)
{
  char tmpstr[160];

  printf(str);
  scanf("%s", tmpstr);
  printf("\n");
  if (toupper(tmpstr[0]) == 'Y')
    return 'Y';
  return 'N';
}

/*-----------------------------------------------------------------------------
 isa_setup -
|-----------------------------------------------------------------------------*/
int isa_setup(void)
{
 unsigned int addr;
 int stat;
 int done = 0;

  while ((total_boards < 4) && (!done))
  {
    stat = get_num("Enter an io-address for this RocketPort board:", &addr,1);
    if (((addr == 0) || (addr > 0x400)) ||
        ((addr & 0x3f) > 0)  )
    {
      printf("This address is not valid.\n");
      printf("Typical addresses include 180, 100, 140, 200, 240,..\n");
    }
    else
    {
      done = 1;
      if (total_boards < 4)
      {
        brds[total_boards].Base_Addr = addr;
        brds[total_boards].num_ports = 0;
        ++total_boards;
        stat = yesno("Do you have more RocketPort ISA boards to setup(y/n)?:");
        if (stat == 'Y')
          done = 0;
      }
    }
  }
  return 0;  // ok
}

/*-----------------------------------------------------------------------------
 pci_setup -
|-----------------------------------------------------------------------------*/
int pci_setup(void)
{
  int num_brds, i;

   for (i=0;i<MAX_PCI_BRDS; i++)
   {
     brds[i].Base_Addr = 0;
     brds[i].Irq = 0;
   }

   aaFindPCIBoards(&brds[0], &num_brds);

   if (num_brds != 0)
   {
     printf("%d PCI boards found\n", num_brds);
     for (i=0; i<num_brds; i++)
     {
       printf("PCI board %d has addr:%xH, IRQ:%d Ports:%d\n", i+1,
         brds[i].Base_Addr,
         brds[i].Irq,
         brds[i].num_ports);
     }

     total_boards = num_brds;
   }
   else
   {
     printf("No PCI RocketPort Boards found.\n");
   }
  return 0;
} 

